''' Entry point which sets a few Kivy configuration parameters before launching the application
'''

import sys
import os

from kivy.logger import Logger
from kivy.app import App
from kivy.core.window import Window
from kivy.core.text import LabelBase
from kivy.properties import NumericProperty, ListProperty

from jocular.components.scope import Scope

class Jocular(App):

    dimmed_transp = NumericProperty(.7)
    ring_transparency = NumericProperty(1)
    text_color = ListProperty([.8, .8, .8, 1])
    knob_on_color = ListProperty([.7, .7, .7, 1])
    knob_color = ListProperty([.3, .3, .3, 1])
    switch_color = ListProperty([.3, .3, .3, 1])
    highlight_color = ListProperty([1, 69/255, 0, 1])
    lowlight_color = ListProperty([.3, .3, .3, 1])
    background_color = ListProperty([.06, .06, .06, 1])
    line_color = ListProperty([.3, .3, .3, 1])

    def on_dimmed_transp(self, *args):
        for c in ['text', 'highlight', 'lowlight', 'knob', 'knob_on', 'switch', 'line']:
            v = getattr(self, c + '_color')
            v[-1] = self.dimmed_transp

    def build(self):
        self.scope = Scope(self, base_path, jocular_version)
        return self.scope

    def on_start(self):
        pass

    def on_stop(self):
        if hasattr(self, 'scope'):
            self.scope.on_close()

if __name__ == "__main__":

    # gets rid of virtual keyboard, multi-touch etc
    from kivy.config import Config
    Config.set('kivy', 'keyboard_mode', 'system')
    Config.set('kivy', 'log_enable', 0)
    Config.set('kivy', 'exit_on_escape', '0')
    Config.set('graphics', 'position', 'auto')
    Config.set('input', 'mouse', 'mouse,multitouch_on_demand')

    if len(sys.argv) > 1:
        if sys.argv[1] in ['trace', 'debug', 'info', 'warning', 'error', 'critical']:
            Config.set('kivy', 'log_level', sys.argv[1])
    else:
        Config.set('kivy', 'log_level', 'error')

    Config.write()
    Config.get('kivy', 'keyboard_mode')
    Config.get('kivy', 'log_level')

    jocular_version = '0.0.1'
    
    try:
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath('.')

    LabelBase.register(name='Jocular', fn_regular=os.path.join(base_path, 'jocular', 'resources', 'jocular4.ttf'))

    # find screen centre in order to centre on new size
    l, t = Window.left, Window.top
    w, h = Window.size
    cx, cy = l + w // 2, t + h // 2

    # set to desired size if possible, but ensure no wider/higher than screen size
    width, height = min(2 * cx, 1280), min(2 * cy - 40, 800)

    Window.size = width, height
    Window.left = max(0, cx - width // 2)
    Window.top = max(0, cy - height // 2)

    try:
        app = Jocular()
        app.run()
        
    except Exception as e:
        Logger.exception('Jocular: root exception!')
        app.on_stop()
        sys.exit('Jocular failed: {:}'.format(str(e)))
