'''Aligner performs star extraction, centroid estimation and sub alignment.
Called from Sub.process in order to support incremental recompute.

Future plans:
* support realign from keyframe
* handle alignment of stacked images for multispectral
* FWHM estimation and other measures of sub quality
* make this into more of a utility library rather than a component of Jocular
* move some methods out (eg centroiding) as generally useful for photometry
'''

import warnings
import numpy as np

from skimage.measure import moments, ransac
from skimage.transform import EuclideanTransform, matrix_transform
from skimage.feature import blob_dog

from kivy.logger import Logger
from kivy.properties import BooleanProperty, NumericProperty, OptionProperty
from kivy.event import EventDispatcher

from jocular.utils.component import Component

class Aligner(EventDispatcher, Component):

    do_align = BooleanProperty(True)

    # star extraction
    adapt_star_threshold = BooleanProperty(True)
    star_intensity = NumericProperty(1)
    min_sigma = NumericProperty(3)
    max_sigma = NumericProperty(5)
    min_stars = NumericProperty(3)

    # centroiding
    centroid_radius = NumericProperty(8)
    centroid_method = OptionProperty('full', options=['full', 'none', 'fast'])

    # registration via RANSAC
    ransac_min_samples = NumericProperty(5)
    ransac_residual_threshold = NumericProperty(.5)
    ransac_max_trials = NumericProperty(100)
    min_inliers = NumericProperty(3)

    centroid_radius = NumericProperty(8)

    def __init__(self, scope):

        Logger.debug('Aligner: initialising')
        self.settables = self.properties()
        self.scope = scope
        super().__init__()
        self.load_settings(settings=scope.settings)

    def post_init(self, *args):
        self.on_centroid_radius()

    # prepare for new alignment
    def reset(self):
        Logger.debug('Aligner: resetting')
        self.keystars = None
        self.forward_warp_model = None
        self.backward_warp_model = None

    # on any event that needs a recompute, call check_recompute
    def on_star_intensity(self, *args):     self.scope.stacker.check_recompute()
    def on_do_align(self, *args):           self.scope.stacker.check_recompute()
    def on_min_sigma(self, *args):          self.scope.stacker.check_recompute()
    def on_max_sigma(self, *args):          self.scope.stacker.check_recompute()
    def on_min_stars(self, *args):          self.scope.stacker.check_recompute()
    def on_centroid_method(self, *args):    self.scope.stacker.check_recompute()
    def on_ransac_min_samples(self, *args): self.scope.stacker.check_recompute()
    def on_ransac_residual_threshold(self, *args): self.scope.stacker.check_recompute()
    def on_ransac_max_trials(self, *args):  self.scope.stacker.check_recompute()
    def on_min_inliers(self, *args):        self.scope.stacker.check_recompute()

    def on_centroid_radius(self):
        # precompute pixel grid and mask for evaluating centroids
        r = self.centroid_radius
        patch_size = 2 * r + 1
        self.grid_x, self.grid_y = np.meshgrid(np.arange(patch_size), np.arange(patch_size))
        self.mask = np.sqrt((r - self.grid_x)**2 + (r - self.grid_y)**2) <= r

    def align(self, centroids, align_direction='forward'):
        '''Align im to current keystars, returning a pair of value warp_model, success
        If we have no keystars so far and hence no warp model, warp_model is None
        If we cannot perform alignment, return False in success

        align_direction controls whether are aligning forwards or backwards
        from a keyframe. We maintain warp model in each direction to
        project stars to predicted positions prior to actual alignment
        step. This will support realignment from a different keyframe in a 
        future version of the method.
        '''

        # if no keystars yet, save centroids as keystars
        if self.keystars is None:
            Logger.debug('Aligner: keyframe')
            self.keystars = centroids
            return None, None
        
        # choose warp model to project existing keystars so they lie closer to centroids (in theory)
        keys = self.keystars
        if (align_direction == 'forward') and self.forward_warp_model:
            keys = matrix_transform(self.keystars, self.forward_warp_model.params)
        elif (align_direction == 'backward') and self.backward_warp_model:
            keys = matrix_transform(self.keystars, self.backward_warp_model.params)
        else:
            keys = self.keystars

        # find closest matching star to each transformed keystar
        matched_stars = np.zeros((len(keys), 2))
        for i, (x1, y1) in enumerate(keys):
            matched_stars[i, :] = centroids[np.argmin([(x1 - x2) ** 2 + (y1 - y2) ** 2 for x2, y2 in centroids])]

        # apply RANSAC to find which matching pairs best fit Euclidean model
        # can throw a warning which we ignore
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            warp_model, inliers = ransac(
                (np.array(self.keystars), matched_stars),
                EuclideanTransform,
                self.ransac_min_samples, 
                self.ransac_residual_threshold, 
                max_trials=self.ransac_max_trials)

        success = (inliers is not None) and (sum(inliers) >= self.min_inliers)

        if success:
            # update forward or backward warp model
            if align_direction == 'forward':
                self.forward_warp_model = warp_model
            else:
                self.backward_warp_model = warp_model
        else:
            Logger.info('Aligner: insufficient inliers')

        return warp_model, success

    def extract_stars(self, im, first_sub=False):
        '''Extracts star coordinates. If we don't find enough or too many, change the
        star threshold and start again until we run into hard limits. Then return no
        stars and inform user.

        Note this is currently potentially doing a lot of repeated work on the first sub
        Better approach would be to implement DOG locally then simply apply the star threshold
        search to the resulting filtered image. That way we can also pick up the real pixel intensities
        rather than the filtered versions.

        This should become in future a proper searh, not this fairly arbitrary approach. A better idea would be to 
        look for bimodality in the effect on nstars as a function of star threshold. For low thresholds that 
        are really picking up background noise, we should find a fairly fast changeover point where
        the number of stars drops rapidly with increasing threshold. 
        '''

        stars = blob_dog(im, 
            min_sigma=self.min_sigma, 
            max_sigma=self.max_sigma, 
            threshold=self.star_intensity / 100, 
            overlap=0)[:, [1, 0]]

        # if we are adapting, only do it on FIRST sub
        if (self.adapt_star_threshold) & first_sub:
            if len(stars) > 200:
                # bracket any changes with don't check flag to prevent wasted work when star_intensity changes
                self.scope.stacker.dont_check_recompute = True
                self.star_intensity *= 1.5
                self.scope.stacker.dont_check_recompute = False
                if self.star_intensity < 10:
                    Logger.debug('Aligner: increased min star intensity to {:5.2f}'.format(self.star_intensity))
                    return self.extract_stars(im, first_sub=True)
                else:
                    Logger.debug('Aligner: Star threshold too high to extract stars')
                    return []
            elif len(stars) < self.min_stars:
                self.scope.stacker.dont_check_recompute = True
                self.star_intensity *= .8
                self.scope.stacker.dont_check_recompute = False
                if self.star_intensity > .1:
                    Logger.debug('Aligner: decreased min star intensity to {:5.2f}'.format(self.star_intensity))
                    return self.extract_stars(im, first_sub=True)
                else:
                    Logger.debug('Aligner: Star threshold too low to extract stars')
                    return []

        return stars

    def compute_centroids(self, im, stars):
        # extract info from all candidate stars and return centroids

        method = self.centroid_method
        if method == 'none':
            return stars
 
        r = self.centroid_radius
        rows, cols = im.shape
        star_data = np.zeros((len(stars), 2))
        ptr = 0
        for x, y in stars:
            x, y = int(x), int(y)
            if (x >= r) and (y >= r) and (x < (cols - r)) and (y < (rows - r)):
                imfrag = im[(y - r):(y + r + 1), (x - r):(x + r + 1)].copy()
                if method == 'full':
                    cx, cy = self.compute_centroid(imfrag)
                    if cx > 0:
                        star_data[ptr, :] = [cx + x - r, cy + y - r]
                        ptr += 1
                elif method == 'fast':
                    cx, cy = self.fast_compute_centroid(imfrag)
                    star_data[ptr, :] = [cx + x - r, cy + y - r]
                    ptr += 1

        return star_data[:ptr, :]

    def compute_centroid(self, imfrag):
        # slightly more sophisticated approach
        
        # mean intensity within masked circle
        mean_intensity = np.mean(imfrag[self.mask])
        
        # star pixels are in mask and above mean intensity
        candidates = self.mask & (imfrag > mean_intensity)
        
        if candidates.any():
            # estimate background and subtract from image
            mean_background = np.mean(imfrag[np.logical_not(candidates)])
            imfrag_sub = candidates * (imfrag - mean_background)       
            wim = np.sum(imfrag_sub)
            cx = np.sum(self.grid_x * imfrag_sub) / wim
            cy = np.sum(self.grid_y * imfrag_sub) / wim
            return cx, cy
        else:
            return -1, -1

    def fast_compute_centroid(self, im):
        # naive but faster approach
        M = moments(im, order=1)
        cy = M[1, 0] / M[0, 0]
        cx = M[0, 1] / M[0, 0]     
        return cx, cy

