# Handles all details of calibration, including dark library and bad pixel map

import os.path
import json
import shutil
import numpy as np
from datetime import datetime

from scipy.ndimage import convolve
from astropy.io import fits 

from kivy.logger import Logger
from kivy.properties import OptionProperty, BooleanProperty, NumericProperty, StringProperty, DictProperty
from kivy.event import EventDispatcher
from kivy.core.window import Window

from jocular.widgets.table import Table
from jocular.utils.utils import add_if_not_exists, remove_nulls_from_dict, date_time_format, hours_since_date
from jocular.utils.component import Component

class Calibrator(EventDispatcher, Component):

    policies = ['originals', 'closest to capture', 'most recent']

    apply_BPM = BooleanProperty(True)
    bpm_frames = NumericProperty(3)             # num frames to use in BPM

    min_subs_for_master = NumericProperty(7)    # need at least this many to create master
    masters = DictProperty({})
    apply_dark = BooleanProperty(False)
    temperature_tolerance = NumericProperty(5)
    exposure_tolerance = NumericProperty(5)
    dark_policy = OptionProperty('closest to capture', options=policies)

    current_dark = StringProperty('')
    current_flat = StringProperty('')
    current_bias = StringProperty('')

    apply_flat = BooleanProperty(False)
    flat_policy = OptionProperty('originals', options=policies)

    bias_policy = OptionProperty('closest to capture', options=policies)

    settables = ['temperature_tolerance', 'exposure_tolerance', 
            'apply_BPM', 'apply_dark', 'apply_flat', 'dark_policy', 'flat_policy', 
            'bias_policy', 'bpm_frames', 'min_subs_for_master']
    
    def __init__(self, scope):

        Logger.debug('Calibrator: initialising')
        self.scope = scope
        super().__init__()
        self.load_settings(settings=scope.settings)
        self.master_fits = {}  # maps type to pair of name of fits, fits data (supports lazy loading)
        self.load_library()
        self.load_bad_pixel_map()

    def build_calibrations(self):

        return Table(
            size=Window.size,
            data=self.library,
            name='Calibration masters',
            cols={
                'Name': {'w': 1, 'align': 'left', 'field': 'master'},
                'Type': {'w': 50, 'field': 'sub_type'},
                'Exposure': {'w': 80, 'field': 'exposure'},
                'Temperature': {'w': 120, 'field': 'temperature'},
                'Filter': {'w': 80, 'field': 'filter'},
                'Created': {'w': 160, 'field': 'created', 'sort': {'DateFormat': date_time_format}},
                'Size': {'w': 120, 'field': 'dims'},
                'Num subs': {'w': 120, 'field': 'n_subs', 'type': int}
                },
            actions={'move to delete folder': self.move_to_delete_folder},
            on_hide_method=self.scope.table_hiding
            )

    def post_init(self, *args):
        self.update_masters()

    def show_calibration_table(self, *args):
        '''Called from menu'''

        if not hasattr(self, 'calibration_table'):
            self.calibration_table = self.build_calibrations()
        self.scope.showing = 'calibration'

        # check for redraw
        if self.calibration_table not in self.scope.children:
            self.scope.add_widget(self.calibration_table, index=0)

        self.calibration_table.show()    

    def on_close(self):
        self.save_bad_pixel_map()
        self.save_library()

    def move_to_delete_folder(self, *args):
        for s in self.calibration_table.selected:
            if s in self.library:
                self._delete(self.library[s]['master'] + '.fit')
                del self.library[s]
        self.calibration_table.update()

    def _delete(self, nm):
        to_path = self.scope.deleted_masters_dir
        Logger.debug('Calibrator: deleting master {:}'.format(nm))
        add_if_not_exists(to_path)
        to_path = os.path.join(to_path, nm)
        if os.path.exists(to_path):
            to_path += datetime.now().strftime('%d_%b_%y_%H_%M_%S')
        Logger.debug('Calibrator: moved to {:}'.format(to_path))
        self.scope.notifier.notify('Moved master to {:}'.format(to_path))
        from_path = os.path.join(self.scope.calibration_dir, nm)
        try:
            shutil.move(from_path, to_path)
        except Exception as e:
            Logger.exception('Calibrator: _delete')

    def on_temperature_tolerance(self, *args):  self.update_masters()
    def on_exposure_tolerance(self, *args):     self.update_masters()
    def on_dark_policy(self, *args):            self.update_masters()
    def on_flat_policy(self, *args):            self.update_masters()
    def on_bias_policy(self, *args):            self.update_masters()
    def on_apply_dark(self, *args):             self.update_masters()
    def on_apply_flat(self, *args):             self.update_masters()
    def on_bpm_frames(self, *args): self.scope.stacker.check_recompute()
    def on_apply_BPM(self, *args):  self.scope.stacker.check_recompute()

    def calibrate_flat(self, im):
        # normalise brightness of flats, and divide by 2 so B & W controls work; 
        # we will reinstate this by doubling prior to saving the master
        # this approach may need to change e.g. computing intensity after subtracting
        # flat darks (or bias) and in the middle of the image
        # if bias is available, subtract it
        # ideally this should be a flat dark
        if 'bias' in self.masters:
            b = self.get_master('bias')
            if b is not None:
                im = im - b

        # normalise by mean of image in central zone
        zone = 100
        h, w = im.shape
        lox, hix = int(h/2 - zone), int(h/2 + zone)
        loy, hiy = int(w/2 - zone), int(w/2 + zone)

        return .5 * im / np.mean(im[lox:hix, loy:hiy])

    def calibrate_light(self, im):
        # lazy load masters

        do_dark = 'dark' in self.masters
        do_flat = 'flat' in self.masters
        do_bias = 'bias' in self.masters

        if do_dark:
            d = self.get_master('dark')
            if d is None:
                do_dark = False
                del self.masters['dark']
        if do_flat:
            f = self.get_master('flat')
            if f is None:
                do_flat = False
                del self.masters['flat']
        if do_bias:
            b = self.get_master('bias')
            if b is None:
                do_bias = False
                del self.masters['bias']

        if not any([do_bias, do_dark, do_flat]):
            return im

        # check shapes in case of eg binning
        if do_dark and (d.shape != im.shape):
            Logger.debug('Calibrator: different shapes')
            return im
        if do_flat and (f.shape != im.shape):
            Logger.debug('Calibrator: different shapes')
            return im
        if do_bias and (b.shape != im.shape):
            Logger.debug('Calibrator: different shapes')
            return im

        # only use bias if we have flats
        if do_bias and (not do_flat):
            do_bias = False

        doing = ['dark' if do_dark else '', 'flat' if do_flat else '', 'bias' if do_bias else '']
        Logger.debug('Calibrator: {:}'.format(' '.join([d for d in doing if d])))

        # We handle all 8 cases even though they might not make complete sense

        # this value might need to be changed for different sensors
        lower_limit = 1 / 2 ** 16

        # check if given that the flat is scaled with a mean of 1 whether any need to subtract bias
        # big diff between im - b/f - b and im/f

        if do_dark and do_flat and do_bias:
            #im = (im - d) / np.maximum(lower_limit, (f - b))
            im = (im - d) / (f - b)

        # no darks but flats and bias
        elif do_flat and do_bias:
            im = (im - b) / np.maximum(lower_limit, (f - b))
            #im = (im - b) / np.maximum(lower_limit, f)

        # no bias: use dark as synthetic bias
        elif do_dark and do_flat:
            im = (im - d) / np.maximum(lower_limit, (f - d))

        # no flats: ignore bias
        elif do_dark and do_bias:
            im = im - d

        # only darks
        elif do_dark:
            im = im - d

        # only flats: won't produce good results
        elif do_flat:
            im = im / np.maximum(lower_limit, f)

        # only bias    
        elif do_bias:
            im = im - b

        # reapply BPM at this point based on saved map (quick) to deal with any artefacts due to incomplete BPM?
        im = self.do_bpm(im)

        # clip since division could result in values > 1
        return np.clip(im, 0, 1)

    def get_master(self, typ):
        try:
            reload = True    
            nm = self.masters[typ]
            if typ in self.master_fits:
                reload = self.master_fits[typ][0] != nm
            if reload:
                Logger.debug('Calibrator: loading master {:}'.format(nm))
                d = fits.open(os.path.join(self.scope.calibration_dir, nm + '.fit'))[0].data
                if typ == 'dark':
                    d = self.do_bpm(d, self.find_hot_pixels(d))  # remove hotpix based on dark itself
                self.master_fits[typ] = nm, d
            return self.master_fits[typ][1]
        except Exception as e:
            Logger.exception('Calibrator: problem loading master')
            return  None

#===============================================
# load and save calibration masters  and library
#===============================================

    def load_library(self):
        try:
            Logger.debug('Calibrator: loading library')
            with open(os.path.join(self.scope.calibration_dir, 'calibration_library.json'), 'r') as f:
                self.library = json.load(f)
            # Flag whether masters exist
            for k, v in self.library.items():
                self.library[k]['exists'] = os.path.exists(os.path.join(self.scope.calibration_dir, v['master'] + '.fit'))
        except Exception as e:
            Logger.info('Calibrator: creating new calibration library')
            self.library = {}

    def save_library(self):
        if not hasattr(self, 'library'):
            return
        Logger.debug('Calibrator: saving library')
        try: 
            with open(os.path.join(self.scope.calibration_dir, 'calibration_library.json'), 'w') as f:
                json.dump(self.library, f, indent=2)
        except Exception as e:
            Logger.exception('Calibrator: save_library')

    def update_masters(self):
        '''Called on changes that could conceivably affect choice of masters
        '''

        def _closest(cands, to_date=None):
            # select candidates closest to the given date; if none, choose most recent

            if not cands:
                return None

            if not to_date:
                to_date = None

            # get create dates as datetimes
            tims = {k: datetime.strptime(v['created'], date_time_format) for k, v in cands.items()}

            if to_date is None:
                # choose most recent
                chosen = [k for k, v in sorted(tims.items(), reverse=True, key=lambda x: x[1])][0]
            else:
                target_time = datetime.strptime(to_date, date_time_format)
                tdeltas = {k: np.abs(t - target_time) for k, t in tims.items()}
                chosen = [k for k, v in sorted(tdeltas.items(), key=lambda x: x[1])][0]

            return cands[chosen]['master']

        def _find_candidates(typ=None):
            '''Find masters of correct type. TO DO: implement selection by filter for flats
            '''

            # prune down to correct type and shape
            cands = {k: v for k, v in self.library.items() 
                if (v['sub_type'] == typ) & 
                    (self.current_image_shape == v['dims']) &
                    v['exists']}

            # return if no cands
            if not cands:
                return cands

            if typ in ['bias', 'flat']:  # 
                return cands

            # rest of method handles darks

            # If exposure is not known (0), don't do anything since recompute will handle it later
            if (self.scope.obj.exposure == 0):
                return

            # If user hasn't set temperature, expand temperature tolerance but restrict time tolerance
            # in order to pick up within-session darks
            if self.scope.obj.temperature_inactive:
                temp_tol = 100
                time_tol = 12  # hours, ie catch anything in current session
            else:
                temp_tol = self.temperature_tolerance
                time_tol = 99999

            # look for darks within temp, exposure and time tolerance. If more than one, choose best score
            # (sum of temp and exposure diffs). Return all those with equal best scores.
            # now = datetime.now()
            scores = {}
            for k, v in cands.items():
                temp_diff = np.abs(v['temperature'] - self.scope.obj.temperature)
                exp_diff = np.abs(v['exposure'] - self.scope.obj.exposure)
                time_diff = hours_since_date(v['created'])
                if (temp_diff <= temp_tol) and \
                    (exp_diff <= self.exposure_tolerance) and \
                    (time_diff < time_tol):
                    scores[k] = exp_diff + temp_diff

            if scores:
                minscore = min(list(scores.values()))
                return {k: v for k, v in cands.items() if (k in scores) and (scores[k] == minscore)}
            else:
                return {}

        #---------------------------
        # main part of method
        #---------------------------

        # no calibration masters, so return
        self.masters = {}
        if (not hasattr(self, 'library')) or (not self.library):
            return

        # don't update if we have no object loaded yet
        if not hasattr(self.scope, 'stacker'):
            return
        else:
            if self.scope.stacker.is_empty():
                return

        # get current stack image shape
        self.current_image_shape = self.scope.stacker.shape()
        if self.current_image_shape is None:
            return

        Logger.debug('Calibrator: updating choice of masters shape {:}'.format(self.current_image_shape))

        master_types_needed = []
        if self.apply_dark:
            master_types_needed += ['dark']
        if self.apply_flat:
            master_types_needed += ['flat', 'bias']

        # apply policy for offline mode
        if not self.scope.live:
            dark_policy = self.dark_policy
            flat_policy = self.flat_policy
            bias_policy = self.bias_policy
            session_date = self.scope.obj.session
        else:
            dark_policy = flat_policy = bias_policy = 'most recent'
            session_date = None
    
        for typ in master_types_needed:
            nm = None
            policy = getattr(self, typ + '_policy')
            try:
                if policy == 'originals':
                    if typ in self.scope.obj.original_masters:
                        nm = self.scope.obj.original_masters[typ]
                elif policy == 'most recent':
                    nm = _closest(_find_candidates(typ))
                elif policy == 'closest to capture':
                    nm = _closest(_find_candidates(typ), to_date=session_date)
            except Exception as e:
                Logger.exception('Calibrator: update_masters')
                nm = None

            # update master fits and masters available
            if nm is not None:
                self.masters[typ] = nm

        self.current_dark = self.masters['dark'] if 'dark' in self.masters else ''
        self.current_flat = self.masters['flat'] if 'flat' in self.masters else ''
        self.current_bias = self.masters['bias'] if 'bias' in self.masters else ''
 
        Logger.debug('Calibrator: using {:}'.format(self.masters.keys()))

        # check if we need to recompute based on change in masters
        self.scope.stacker.check_recompute()

    def save_master(self, master_name=None):
        '''Save master, constructed from current stack. Called from Obj save. If master_name is
        given, use this (ie overwrite if necessary); otherwise invent one.
        Return master name
        '''

        stacker = self.scope.stacker
        nsubs = len(stacker.subs)

        if nsubs < self.min_subs_for_master:
            Logger.debug('Calibrator: insufficient subs {:2d}'.format(nsubs))
            self.scope.notifier.notify('Insufficient subs {:2d} to create master.'.format(nsubs))
            return None

        Logger.debug('Calibrator: saving master {:}'.format(master_name))

        sub_type = self.scope.obj.sub_type

        # compute final stack
        stacker.force_update = True
        master = stacker.compute_stack()
        stacker.force_update = False

        # flats are divided thru by their median to account for level differences during
        # acquisition, but we scaled them to 50% to enable B/W controls to make sense; we
        # therefore now multiply by 2 prior to saving
        if sub_type == 'flat':
            master = 2 * master

        created = datetime.now().strftime(date_time_format)
        created_ = datetime.now().strftime('%d_%b_%y_%H_%M_%S')

        # for the future
        # filt = stacker.subs[0].filt
        filt = 'L'
        
        # current exposure & temperature
        exposure = self.scope.obj.exposure
        temperature = self.scope.obj.temperature

        # number of selected subs
        n_subs = len([s for s in stacker.subs if s.status == 'select'])

        # generate master name for appropriate type
        if (master_name is None) or (len(master_name) == 0):
            if sub_type == 'flat':
                master_name = 'flat_{:s}_{:3.0f}s_{:s}'.format(filt, exposure, created_)
            elif sub_type == 'bias':
                master_name = 'bias_{:3.0f}s_{:s}'.format(exposure, created_)
            elif sub_type == 'dark':
                master_name = 'dark_{:d}C_{:3.0f}s_{:s}'.format(temperature, exposure, created_)

        master_name = master_name.replace(' ','')

        name = '{:} {:}'.format(sub_type, created_) 

        # if master is already in library, remove it
        loc = [k for k, v in self.library.items() if v['master'] == master_name]
        if loc:
            del self.library[loc[0]]

        hdr = {
            'sub_type': sub_type,
            'n_subs': n_subs,
            'exposure': exposure,
            'temperature': temperature,
            'filter': filt,
            'created': created,
            'master': master_name,
            'dims': [master.shape[0], master.shape[1]],
            'exists': True
        }

        self.library[name] = remove_nulls_from_dict(hdr) 

        # save
        try:
            # save master
            hdu = fits.PrimaryHDU(master)
            hdu.writeto(os.path.join(self.scope.calibration_dir, master_name + '.fit'), overwrite=True)
            Logger.debug('Calibrator: saved as {:}'.format(master_name))

            # save to calibration library
            self.save_library()

            # update masters to capture most recent
            self.update_masters()

        except Exception as e:
            Logger.exception('Calibration: save_master')
            return None

        return master_name

#-------------------
# bad pixel mapping 
#-------------------

    def process_bpm(self, im):
        if self.apply_BPM:
            self.update_bpm(self.find_hot_pixels(im))
            return self.do_bpm(im, self.bpm)
        else:
            return im

    def find_hot_pixels(self, im):
        '''Return hot pixel candidates.
        We look for non-edge pixels whose intensity is significantly greater
        than their neighbours. Approach is conservative in order to
        find all hot pixels, since (a) we remove any imposters by comparing
        across most recent 'bpm_frames' when building the hot pixel map; and (b) in the
        EAA use case the only adverse effect is to replace a few non-hot pixels
        by their median. Returns a set of (row, col) coordinates
        ''' 
                
        # set min to zero
        im_norm = im - np.min(im)

        # divide im by local sum in 3x3 region
        im2 = im_norm / convolve(im_norm, np.ones((3, 3)), mode='constant')

        # Define hot pix as more than 5 SD from mean
        hp_cands = (im2 > np.mean(im2) + 5*np.std(im2))

        # set boundaries to zero
        hp_cands[0, :] = 0
        hp_cands[-1, :] = 0
        hp_cands[:, 0] = 0
        hp_cands[:, -1] = 0

        # coordinates of hot pixels as 2 x N array
        hps = np.where(hp_cands>0)

        # return hot pixel coordinates as a set of row-col pairs
        return {(r, c) for r, c in zip(hps[0], hps[1])}

    def do_bpm(self, im, bpm=None):
        '''Replace each pixel in bad pixel map by median of neighbours'''
        if bpm is None:
            bpm = self.bpm
        if bpm is not None:
            for r, c in bpm:
                im[r, c] = np.median(im[r-1:r+2, c-1:c+2].ravel())
        return im
 
    def load_bad_pixel_map(self):
        self.bplist = []  # acts as a circular buffer of BPs from recent frames
        self.frame_count = 0
        self.bpm = None
        bpfile =  os.path.join(self.scope.calibration_dir, 'bad_pixel_map.npy')
        if os.path.exists(bpfile):
            bpm = np.load(bpfile)
            self.update_bpm({(x, y) for x, y in zip(bpm[0], bpm[1])})

    def compute_bpm(self):
        # Use intersection of bad pixels from previous N frames to compute bad pixel map
        if self.bplist:
            self.bpm = self.bplist[0]
            for bpl in self.bplist[1:]:
                self.bpm = self.bpm.intersection(bpl)

    def update_bpm(self, bpm):
        '''Add new BPM set to BPM, recomputing BPM'''
        if len(self.bplist) == self.bpm_frames:
            self.bplist[self.frame_count % self.bpm_frames] = bpm
        else:
            self.bplist.append(bpm)
        self.frame_count += 1
        self.compute_bpm()

    def save_bad_pixel_map(self):
        '''Save as npy file'''
        if self.bpm is None:
            return
        bpm = np.array([[x for x, y in self.bpm], [y for x, y in self.bpm]])
        bpfile =  os.path.join(self.scope.calibration_dir, 'bad_pixel_map.npy')
        np.save(bpfile, bpm)


