'''camera.py 
Simple camera simulator to test live observing. Works by looking for images in the light, dark, flat or bias folders
within the simulator folder of captures. If any found, they are delivered to the watched folder at the specified
exposure rate.
'''

import os
import shutil

from kivy.logger import Logger
from kivy.properties import OptionProperty, ObjectProperty
from kivy.event import EventDispatcher
from kivy.clock import Clock

from jocular.utils.component import Component
from jocular.utils.utils import get_fits

class Camera(EventDispatcher, Component):

    camera = OptionProperty('simulator', options=['simulator'])
    capture_event = ObjectProperty(None, allownone=True)

    def __init__(self, scope):

        Logger.debug('Camera: initialising')
        self.settables = ['camera']
        super().__init__()
        self.load_settings(settings=scope.settings)
        self.scope = scope

    def is_capturing(self):
        return self.capture_event is not None

    def start(self, sub_type, *args):

        Logger.debug('Camera: starting capture')

        # in case this is called programatically and capture is still ongoing
        if self.is_capturing():
            self.capture_event.cancel()

        # names of subs of right type in simulation dir (if any), in time order 
        self.subs = get_fits(os.path.join(self.scope.simulator_dir, sub_type), time_order=True)
        self.count = 0
        exposure = max(1, self.scope.obj.exposure) # don't allow short exposures at present
        self.capture_event = Clock.schedule_interval(self.generate_sub, exposure)

    def generate_sub(self, *args):
        '''Dump next sub of current type into watched folder
        '''
        if self.count < len(self.subs):
            shutil.copy(self.subs[self.count], self.scope.watched_dir)
            self.count += 1
        else:
            self.stop()

    def stop(self):
        if self.is_capturing():
            Logger.debug('Camera: stopping capture')
            self.capture_event.cancel()
            self.capture_event = None

    def on_close(self, *args):
        self.stop()

 