# Configurer. Handles user setting of configuration options.

import os
import json
from functools import partial

from kivy.logger import Logger
from kivy.uix.boxlayout import BoxLayout
from kivy.properties import BooleanProperty, OptionProperty, ObjectProperty, StringProperty
from kivy.uix.label import Label
from kivy.lang import Builder
from kivy.metrics import dp

from jocular.utils.component import Component
from jocular.widgets.hyperknob import HyperKnob
from jocular.widgets.jfilebrowser import JFileBrowser

Builder.load_string('''

<Parameter>:
    size_hint: 1, None
    height: dp(19)
    mybut: _button

    ToggleButton:
        id: _button
        group: 'fred'
        size_hint: 1, 1
        font_size: '15sp'
        text_size: self.size
        halign: 'center'
        text: '{:} ({:})'.format(root.title, root.current_value)
        color: app.text_color if self.state == 'normal' else app.highlight_color
        background_color: app.background_color
        shorten: True
        background_normal: ''

<Configurer>:
    size_hint: None, None
    width: root.scope.obj.width
    height: dp(400)
    parameters: _parameters
    settings: _settings
    subtitle_label: _subtitle

    canvas.before:
        Color:
            rgba: app.background_color
        Rectangle:
            pos: self.pos
            size: self.size

    BoxLayout:
        orientation: 'vertical'

        ScrollView:
            size: self.size

            StackLayout:
                size_hint: 1, None
                id: _parameters

        FloatLayout:
            id: _settings
            size_hint: 1, 1

        JLabel:
            id: _subtitle
            size_hint: 1, None
            text_size: self.size
            height: dp(70)
            font_size: '15sp'
            text: root.subtitle
            padding: dp(10), dp(10)
            valign: 'top'
''')


class Parameter(BoxLayout):
    title = StringProperty('title')
    current_value = ObjectProperty(None, allow_none=True)

class Configurer(BoxLayout, Component):

    pinned = BooleanProperty(False)  # used for temporary pinning in this case
    showing = BooleanProperty(False)
       
    configuration = OptionProperty('dummy', 
        options=['dummy', 'Capture', 'Calibration', 'Alignment', 'Display', 'General'],
        allownone=True)

    subtitle = StringProperty('')

    def __init__(self, scope=None, **kwargs):
        self.scope = scope
        super().__init__()
        self.pos = 0, -self.scope.height
        self.showing = False
        try:
            with open(os.path.join(self.scope.base_path, 'jocular', 'config.json'), 'r') as f:
                self.configs = json.load(f)
        except OSError as e:
            Logger.exception('Configurer: missing config.json')
            raise

        # needed for scrolling to work
        self.parameters.bind(minimum_height=self.parameters.setter('height'))

    def redraw(self, *args):
        self.scope.add_widget(self) # , index=1)
        self.size = self.scope.obj.width, self.scope.height - self.scope.processing.height
        self.pos = 0, -self.scope.height
        self.showing = False

    # ensure no clicks get passed thru
    def on_touch_down(self, touch):
        handled = super().on_touch_down(touch)
        if self.collide_point(*touch.pos):
            return True
        return handled

    def set_value(self, v, param_label, *args):
        self.set_parameter_value(v, self.inputter.value)
        param_label.current_value = v['format'].format(self.get_parameter_value(v))

    def set_boolean_value(self, v, param_label, *args):
        self.set_parameter_value(v, self.inputter.selected)
        param_label.current_value = v['format'].format(self.get_parameter_value(v))

    def toggle_parameter(self, v, current_value, param_label):
        new_val = not current_value
        self.set_parameter_value(v, new_val)
        param_label.current_value = v['format'].format(new_val)

    def get_parameter_value(self, v):
        if v['component'] == 'scope':
            return getattr(self.scope, v['name'])
        else:
            return getattr(getattr(self.scope, v['component']), v['name'])

    def set_parameter_value(self, v, value):
        if v['component'] == 'scope':
            setattr(self.scope, v['name'], value)
        else:
            setattr(getattr(self.scope, v['component']), v['name'], value)

    def clear_settings(self, *args):
        self.settings.clear_widgets()
        self.subtitle_label.text = ''

    def get_dirname(self, v, current_value, param_label):
        try:
            self._browser = JFileBrowser(path=os.path.join(current_value, '..'),
                title='Choose or create a directory for monitoring incoming FITs files')
            self.scope.add_widget(self._browser)
            self._browser.bind(
                on_success=self._dirname_chosen,
                on_canceled=self._dirname_cancelled)
            self._v = v
        except OSError:
            Logger.debug('Configurer: unable to set watched folder {:}'.format(os.path.join(current_value, '..')))
            self.scope.notifier.notify('Unable to set watched folder')
        return None

    def _dirname_chosen(self, instance):
        try:
            selection = instance.selection[0]
            if selection:
                self.set_parameter_value(self._v, selection)
            self._dirname_cancelled(instance)
        except:
            Logger.debug('Configurer: folder not changed')
            self._dirname_cancelled(instance)
            self.scope.notifier.notify('Folder not changed')

    def _dirname_cancelled(self, instance):
        self.scope.remove_widget(instance)
 
    def set_variable(self, v, p, *args):

        if 'name' not in v:
            return

        self.settings.clear_widgets()
        self.subtitle = v['Subtitle']

        current_value = self.get_parameter_value(v)

        if v['type'] == 'boolean':
            self.toggle_parameter(v, current_value, p)
            return

        elif v['type'] == 'dirname':
            self.get_dirname(v, current_value, p)
            return

        else:
            factory = v['default']
            if 'options' in v:
                presets = v['options']
                #if type(presets) == dict:
                #    presets = {(k + ', current' if v == current_value else k): v for k, v in presets.items() }
            else:
                presets = {}
                if v['min'] not in [current_value, factory]:
                    presets[v['format'].format(v['min'])] = v['min']
                if v['max'] not in [current_value, factory]:
                    presets[v['format'].format(v['max'])] = v['max']

            if type(presets) == dict:
                presets['current'] = current_value
                if abs(current_value - factory) > .0001:
                    presets['factory'] = factory

            if 'options' in v:
                self.inputter = HyperKnob(
                    pos_hint={'center_x': .5, 'center_y': .5},
                    knob_type='switch',
                    minangle=-120, 
                    maxangle=120,
                    switch_radius=.5,
                    knob_radius=.47,
                    size=(dp(300), dp(300)),
                    presets=presets,
                    value=current_value)
            else:
                self.inputter = HyperKnob(
                    pos_hint={'center_x': .5, 'center_y': .5},
                    knob_type='numeric with presets',
                    minangle=-120, 
                    maxangle=120,
                    min=v['min'], 
                    max=v['max'], 
                    show_value=True,
                    size=(dp(300), dp(300)), 
                    value_format=v['format'],
                    value_font_size='28sp', 
                    switch_radius=.45,
                    knob_radius=.4,
                    value_type=v['type'],
                    presets=presets,
                    value=current_value)

            self.inputter.bind(value=partial(self.set_value, v, p))

        self.settings.add_widget(self.inputter)

    def get_current_val(self, v):
        return v['format'].format(self.get_parameter_value(v))

    def on_configuration(self, *args):

        if self.configuration is None:
            return

        # change current configuration panel
        self.parameters.clear_widgets()
        self.clear_settings()

        if self.configuration not in self.configs:
            Logger.exception('Configurer: no configs entry for {:}'.format(self.configuration))
            return

        confs = self.configs[self.configuration]

        # fill in defaults
        for v in confs:
            if 'format' not in v:
                v['format'] = '{:}'
            if 'Subtitle' not in v:
                v['Subtitle'] = ''
            if 'min' not in v:
                v['min'] = 0
            if 'max' not in v:
                v['max'] = 1
            if 'default' not in v:
                v['default'] = ''
           
        # process booleans first
        bools = [v for v in confs if v['type'] == 'boolean']
        for v in bools:
            p = Parameter(title=v['Title'], current_value=self.get_current_val(v))
            p.mybut.state = 'normal'
            p.mybut.bind(on_press=partial(self.set_variable, v, p))
            self.parameters.add_widget(p)

        if bools:
            # add a blank line
            self.parameters.add_widget(Label(text='', size_hint=(1, None), height=10))

        nonbools = [v for v in confs if v['type'] != 'boolean']
        for v in nonbools:
            p = Parameter(title=v['Title'], current_value=self.get_current_val(v))
            p.mybut.state = 'normal'
            p.mybut.bind(on_press=partial(self.set_variable, v, p))
            self.parameters.add_widget(p)

    # called from scope when enter config panel space; action is to unpin, 
    # to allow it to close on exit
    def enter(self):
        if self.showing:
            self.pinned = False

    def open(self, process):
        self.show()
        if process == self.configuration:
            self.hide()
        else:
            self.configuration = process
            self.pinned = True

    def hide(self, *args):
        if (self.showing) and not (self.pinned):
            self.pos = 0, -self.scope.height
            self.clear_settings()
            self.showing = False
            self.configuration = None

    def show(self, *args):
        if not self.showing:
            self.pos = 0, self.scope.processing.height
            self.showing = True
            self.pinned = True
