''' Eyepiece. Handles the main image display and associated controls. Interface
constructed in Python rather than kv for better control of layers. See redraw for bulk
of construction code.
'''

import numpy as np
from functools import partial

from kivy.logger import Logger
from kivy.vector import Vector
from kivy.graphics.transformation import Matrix
from kivy.core.window import Window
from kivy.graphics import Color, Line
from kivy.uix.floatlayout import FloatLayout
from kivy.properties import (
    BoundedNumericProperty,
    NumericProperty,
    StringProperty, 
    OptionProperty, 
    BooleanProperty, 
    ObjectProperty)
from kivy.clock import Clock
from kivy.metrics import dp

from jocular.utils.gradient import estimate_gradient
from jocular.utils.stretch import stretch
from jocular.utils.utils import angle360, p2cart, joc_icon
from jocular.widgets.lever import Lever
from jocular.widgets.ring import Ring
from jocular.widgets.rotateables import RotLabel, RotatedToggle, RotButton
from jocular.widgets.reticule import Reticule
from jocular.image.imagescatter import ImageScatter
from jocular.utils.component import Component

class Eyepiece(FloatLayout, Component):

    # index of currently selected sub (starting at 0)
    selected_sub = NumericProperty(-1)
    nsubs = NumericProperty(0)
    sub_or_stack = StringProperty('stack')
    scatter = ObjectProperty(None, allownone=True)

    # indicates proportion along stack (range 0-1)
    # stack_position = NumericProperty(1)  # might not be needed

    # object display manipulations
    zoom = NumericProperty(2)
    ring = NumericProperty(1)
    black = NumericProperty(.02)
    autoblack = NumericProperty(-1)
    white = NumericProperty(1)
    p1 = NumericProperty(.2)
    fine = NumericProperty(0)
    fine_sensitivity = NumericProperty(1)
    gradient = NumericProperty(100)
    show_reticule = BooleanProperty(False)
    reset = BooleanProperty(False)
    invert = BooleanProperty(False)
    orientation = BoundedNumericProperty(0, min=0, max=360)
    #binning = NumericProperty(1)
    #binning_on = BooleanProperty(False)
    hypernegative = BooleanProperty(True)
    #apply_ROI = BooleanProperty(False)

    flip_UD = BooleanProperty(False)
    flip_LR = BooleanProperty(False)

    stretch = OptionProperty('hyper', 
        options=['log', 'log γ', 'γ', 'hyper', 'sigmoid', 'asinh', 'linear'])
    stack_combine_method = OptionProperty('mean', options=['mean', 'median', '70%', '80%', '90%'])

    speed = NumericProperty(2)

    redrawing = BooleanProperty(False)

    settables = ['zoom', 'black', 'white', 'p1', 'fine', 
        'gradient', 'show_reticule', 'invert', 'stretch', 'sub_or_stack', 
        'stack_combine_method', 'speed', 'flip_LR', 'flip_UD', 'autoblack', 'hypernegative', 'fine_sensitivity']

    def __init__(self, scope):

        Logger.debug('Eyepiece: initialising')
        self.scope = scope
        super().__init__()
        self.load_settings(settings=scope.settings)
        self.image = None
        self.image_just_updated = False
        self.playing = False

    def reset(self):
        Logger.debug('Eyepiece: reset')
        self.blank_image()
        self.stack = 0
        self.nsubs = 0
        self.selected_sub = -1
        self.update_stack_scroller()

    def blank_image(self):
        # background gradient won't work well with blank image
        self.background_gradient = None
        self.scatter.display_blank()
        self.image = None

    def redraw(self, *args):
        '''Build interface on start and on resize.'''

        def _set_value(name, instance, value):
            setattr(self, name, value)

        def mean(lst):
            return sum(lst) / len(lst)

        Logger.debug('Eyepiece: redrawing')

        self.redrawing = True

        # clear canvas/widgets
        self.clear_widgets()
        self.canvas.clear()

        self.pos = 0, 0
        self.size = Window.height, Window.height  # <-- check this!

        w, h = Window.width, Window.height

        orig = [.5 * (w + self.scope.obj.width) if self.scope.obj.pinned else .5 * w,
                .5 * (h + self.scope.processing.height) if self.scope.processing.pinned else .5 * h
                ]

        # save these, not least for snapshotting the correct region
        self.origin = orig
        self.radius = min(w - orig[0], h - orig[1])

        # create scatter first time thru
        if self.scatter is None:
            self.scatter = ImageScatter(10, 10, eyepiece=self, colorfmt='luminance')
        self.add_widget(self.scatter, index=10)   # send it to the back
        self.scatter._set_center(orig)

        # scope rings: 3 + outer (obscuring) ring        
        inner_width = dp(8)
        middle_width = dp(26)
        outer_width = dp(20)

        radii = {'background': (self.radius - dp(5) , Vector(orig).length())}
        radii['outer'] = radii['background'][0] - outer_width, radii['background'][0]
        radii['middle'] = radii['outer'][0] - middle_width, radii['outer'][0]
        radii['inner'] = radii['middle'][0] - inner_width, radii['middle'][0]
        radii['image'] = 0, radii['inner'][0]

        self.inner_radius = radii['inner'][0]

        greys = {'background': .07, 'outer': .2, 'middle': .17, 'inner': .12, 'image': 0}
        for nm, (r_inner, r_outer) in radii.items():
            th = (r_outer - r_inner) / 2
            if nm != 'image':
                #if nm in ['middle', 'inner', 'outer', 'background']:
                if nm == 'background':
                    orientation_sensor = self
                else:
                    orientation_sensor = None
                my_ring = Ring(pos=orig, radius=r_inner + th, 
                    thickness=th, grey=greys[nm], orientation_sensor=orientation_sensor)
                self.add_widget(my_ring)
                if nm == 'outer':
                    self.outer_ring = my_ring

        # focus/alignment reticule
        self.reticule = Reticule(pos=orig, outer=radii['image'][1], show=self.show_reticule)
        self.add_widget(self.reticule)

        # layout of ring defined here

        lo, hi = 125, 235

        angles = {
            'zoom': [310, 350], 
            'speed': [66, 74], 
            'gradient': [281, 300],
            'ring': [15, 25],
            'black': [hi, lo],
            'white': [hi, lo],
            'fine': [hi, lo],
            'p1': [hi, lo],
            'autoblack': [242, 258],
            'stack_scroller': [100, 80],
            'snapshot': 270,
            'reticule': 0,
            'reset': 45
            }

        lever_color = self.scope.app.highlight_color
        lever_color[-1] = .28

        levers = {
            'zoom': {'values': [.5, 15],  
                'major': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15], 
                'groove': True, 
                'name': 'zoom'},
            'speed': {'values': [0.3, 7], 
                'major': {'slow': .31, 'fast': 6.9},
                'radial': False,
                'name': ''},
            # 'binning': {'values': [.7, 4.1], 
            #     #'major': [1, 2, 3, 4],
            #     'major': {'off': .8, '2': 2, '3': 3, '4': 4}, 
            #     'radial': False,
            #     'name': 'binning'},
            'gradient': {'values': [0, 200], 
                'major': {'off': 0, '-': 30, 'best': 100, '+': 180},
                'groove': True,
                'radial': False,
                'name': 'gradient'},
            'ring': {'values': [-.05, 1.05],
                'major': {'on': 1, 'off': 0},
                'radial': False,
                'name': 'eyepiece',
                'groove': True},
            'black': {'values': [0, 1],
                'marker_len': dp(15),
                #'color': [1, 0, 0, 1],
                'text': 'B',
                'color': self.scope.app.text_color,
                #'color': lever_color,
                #'background_color': [.15, .15, .15, 1],
                'major': [0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1.0], 
                'minor': np.arange(0, 1, .02).tolist(),
                'groove': True},
            'white': {'values': [0, 1],
                'marker_len': dp(15),
                #'background_color': [.15, .15, .15, 1],
                'color': self.scope.app.text_color,
                #'color': lever_color,
                'text': 'W'
                },
            'p1': {'values': [0, 1], 
                'marker_len': dp(20),
                'text': 'S',
                'halign': 'left',
                'color': self.scope.app.text_color,
                #'color': lever_color,
                #'background_color': [.2, .2, .2, 0],
                'radius': radii['background'][0] + dp(10)},
            'fine': {'values': [-.0001, .0001],
                'marker_len': 0,
                'springy': 0, 
                'text': joc_icon('fine'), 
                'radius': radii['inner'][1] + dp(15),
                'background_color': [0, 0, 0, 0],
                'color': lever_color,
                'text_width': dp(30)},
            'autoblack': {'values': [-1, 15], 
                'major': {'off': -1, '-': 2, '+': 14.5}, 
                'radial': False,
                'name': 'auto-black',
                'groove': True}
        }

        radius = radii['middle'][1]
        groove_thickness = outer_width / 2
        name_radius=radii['middle'][0] + middle_width/2
        major_len, minor_len = dp(6), dp(3)

        self.levers = {}  # store the widgets so we can update values on reloading objs
        for name, props in levers.items():

            angs = angles[name]

            props['angles'] = angles[name]
            if 'groove' in props:
                min_a, max_a = angle360(90 - props['angles'][0]), angle360(90 - props['angles'][1])
                self.add_widget(Ring(pos=orig, radius=radius + groove_thickness,
                    thickness=groove_thickness, grey=.15,
                    start_angle=angle360(min(min_a, max_a) - 1), 
                    end_angle=angle360(max(min_a, max_a) + 1)))

            if 'radius' not in props:
                props['radius'] = radii['middle'][1] + outer_width/2

            if 'background_color' not in props:
                props['background_color'] = lever_color

            if 'springy' not in props:
                props['springy'] = None

            lever_props = {k: v for k, v in props.items() if k in 
                {'radius', 'angles', 'values', 'background_color', 'text', 'marker_len',
                    'color', 'text_width', 'springy', 'flatspots'}}

            # create lever widget, initialise, bind so value gets set to property
            lev = Lever(origin=orig, value=getattr(self, name), **lever_props)
            lev.bind(on_value=partial(_set_value, name))
            getattr(self, 'on_{:s}'.format(name))()
            self.add_widget(lev)
            self.levers[name] = lev

            # add decoration
            if 'major' in props:

                if 'radial' not in props:
                    radial = True
                else:
                    radial = props['radial']
                if radial:
                    text_width = dp(20)
                else:
                    text_width = dp(50)

                if type(props['major']) == list:
                    vdict = {str(v): v for v in props['major']}
                else:
                    vdict = props['major']

                with self.canvas:
                    Color(.3, .3, .3, .6)
                    for v in vdict.values():
                        ang = lev.value_to_angle(v)
                        c, s = p2cart(1, ang)
                        x, y = orig[0] + radius*c, orig[1] + radius*s
                        Line(points=(x, y, x - c*major_len, y - s*major_len))

                for vstr, v in vdict.items():
                    ang = lev.value_to_angle(v)
                    self.add_widget(RotLabel(
                        text_width=text_width, font_size='12sp', radial=radial, 
                        text=vstr, angle=ang,
                        color=(.4, .4, .4, .7),
                        radius=radius + dp(10), origin=orig), index=1)

            if 'minor' in props:
                with self.canvas:
                    Color(.4, .4, .4, .7)
                    for v in props['minor']:
                        ang = lev.value_to_angle(v)
                        c, s = p2cart(1, ang)
                        x, y = orig[0] + radius*c, orig[1] + radius*s
                        Line(points=(x, y, x - c*minor_len, y - s*minor_len))

            if ('name' in props) and props['name']:
                self.add_widget(RotLabel(text=props['name'], angle=mean(props['angles']),
                    font_size='12sp', text_width=dp(100), radius=name_radius, origin=orig))

        # representation of stack

        self.sub_colors = {'select': (0, 1, 0, .6), 
            'reject': (.5, .5, .5, .6), 
            'nalign': (1, 0, 0, .6),
            'delete': (.1, .1, .1, .6)}

        fontsizes = [12, 14, 20, 14, 12]
        self.halfwin = (len(fontsizes) - 1) // 2
        self.sub_labels = []
        r = radii['outer'][1] + outer_width/2 - dp(20)
        label_angles = np.linspace(angles['stack_scroller'][0], angles['stack_scroller'][1], 2*self.halfwin + 1)

        for angle, fs in zip(label_angles, fontsizes):
            if fs == fontsizes[self.halfwin]:
                # separate this out because at some point we will implement a delete current sub option
                l = RotButton(text_width=dp(36), text='', angle=float(angle), radius=r, 
                    origin=orig, font_size=fs, valign='middle')
                l.background_color = (0.1, 0.1, 0.1, 1)
                l.height = dp(20)
                l.y -= dp(1)
                l.bind(on_press=partial(self.move_to_sub, l))            
            else:
                l = RotButton(text_width=dp(36), text='', angle=float(angle), radius=r, 
                    origin=orig, font_size=fs, valign='middle')
            #l.bind(on_press=partial(self.move_to_sub, l))
            self.sub_labels.append(l)
            self.add_widget(l)
        self.update_stack_scroller()

        angs = angles['stack_scroller']

        # sub or stack toggle
        self.stack_toggle = RotatedToggle(
            values=['sub', 'stack'],
            text_width=dp(50), 
            text=self.sub_or_stack, 
            angle=mean(angs), 
            radius=radii['middle'][0] + middle_width / 2, 
            origin=orig)
        self.stack_toggle.bind(on_press=self.toggle_sub_or_stack)
        self.add_widget(self.stack_toggle)

        sub_stack_buttons = {'first_sub': angs[0] - 1, 'last_sub': angs[1] + 1, 'prev_sub': angs[0] - 5, 'next_sub': angs[1] + 5}
        for button_name, angle in sub_stack_buttons.items():
            but = RotButton(
                text=joc_icon(button_name, fs=16), 
                text_width=dp(30), 
                angle=angle, 
                radius=radii['middle'][0] + middle_width / 2, 
                origin=orig)
            but.bind(on_press=partial(self.traverse_stack, button_name))
            self.add_widget(but)

        # animation
        play_button = RotatedToggle(
            font_size='20sp', 
            text_width=dp(20), 
            angle=mean(angles['speed']), 
            origin=orig, 
            radial=False,
            radius=radii['middle'][0] + middle_width / 2,
            markup=True,
            text=joc_icon('play stack'))
        play_button.bind(on_press=partial(self.traverse_stack, 'play'))
        self.add_widget(play_button)

        # snapshot button
        snapshot_button = RotatedToggle(
            font_size='20sp', 
            text_width=dp(20), 
            angle=angles['snapshot'], 
            origin=orig, 
            radial=False,
            radius=radii['middle'][0] + middle_width / 2,
            markup=True,
            text=joc_icon('camera', fs=18))
        snapshot_button.bind(on_press=self.scope.snapshotter.save)
        self.add_widget(snapshot_button)

        self.add_widget(RotLabel(
            text='snap', 
            angle=angles['snapshot'],
            font_size='12sp', 
            text_width=dp(50), 
            radius=radii['outer'][0] + outer_width / 2, 
            origin=orig))

        # reticule toggle
        self.reticule_button = RotatedToggle(
            font_size='20sp', 
            text_width=dp(20), 
            angle=angles['reticule'], 
            origin=orig, 
            radial=False,
            radius=radii['middle'][0] + middle_width / 2,
            markup=True,
            text=joc_icon('reticule', fs=20))
        self.reticule_button.bind(on_press=self.on_reticule)
        self.add_widget(self.reticule_button)

        # reset location
        reset_button = RotatedToggle(
            font_size='16sp', 
            text_width=dp(16), 
            angle=angles['reset'], 
            origin=orig, 
            radial=False,
            radius=radii['middle'][0] + middle_width / 2,
            markup=True,
            text=joc_icon('reset', fs=14))
        reset_button.bind(on_press=self.on_reset)
        self.add_widget(reset_button)

        # apply ROI 
        # ROI_button = RotatedToggle(
        #     font_size='32sp', 
        #     text_width=dp(48), 
        #     angle=angles['ROI'], 
        #     origin=orig, 
        #     radial=False,
        #     radius=radii['middle'][0] + middle_width - 4,
        #     markup=True,
        #     text=joc_icon('ROI', fs=18))

        # ROI_button.bind(on_press=self.on_ROI)
        # self.add_widget(ROI_button)

        self.scope.add_widget(self, index=10) 

        self.redrawing = False
        self.on_selected_sub()

    #on_size = redraw

    def update_stack_scroller(self, *args):
        start_sub = self.selected_sub - self.halfwin
        for i, l in enumerate(self.sub_labels, start=start_sub):
            if (i >= 0) and (i < self.nsubs):
                l.text = str(i + 1)
                l.color = self.sub_colors[self.scope.stacker.subs[i].status]
            else:
                l.text = ''

    def traverse_stack(self, action, but, *args):
        if action == 'first_sub':
            self.selected_sub = 0
        elif action == 'last_sub':
            self.selected_sub = self.nsubs - 1
        elif action == 'next_sub':
            if self.selected_sub < self.nsubs - 1:
                self.selected_sub += 1
        elif action == 'prev_sub':
            if self.selected_sub > 0:
                self.selected_sub -= 1
        else:
            if self.playing:
                Clock.unschedule(self.play_event)
                but.text = joc_icon('play stack')
            else:
                self.play_event = Clock.schedule_interval(self.increment_sub, 1/self.speed)
                but.text = joc_icon('pause')
            self.playing = not self.playing

    def increment_sub(self, dt):
        if self.selected_sub == self.nsubs - 1:
            self.selected_sub = 0
        else:
            self.selected_sub += 1

    def move_to_sub(self, sub, *args):
        if len(sub.text) > 0: 
            new_sub = int(sub.text) - 1
            if new_sub == self.selected_sub:
                self.toggle_selected_sub()
            else:                          
                self.selected_sub = max(0, min(new_sub, self.nsubs - 1))
            self.update_stack_scroller()

    def toggle_selected_sub(self):
       # toggle selected
        #new_sub = int(sub.text) - 1
        current_sub = self.selected_sub
        sub = self.scope.stacker.subs[current_sub]
        if sub.status == 'select':
            sub.status = 'reject'
        elif sub.status == 'reject':
            sub.status = 'select'
        self.update_stack_scroller()
        if self.sub_or_stack == 'stack':
            #new_image = self.scope.stacker.compute_stack(up_to=new_sub)
            new_image = self.scope.stacker.compute_stack(up_to=current_sub)
            if new_image is not None:
                self.image = new_image
            self.update_image()

    def sub_added(self):
        # used to update selected_sub on new sub to enable tracking most recent; called by stacker
        self.nsubs = len(self.scope.stacker.subs)
        if self.selected_sub == self.nsubs - 2:  # was on last sub before new one came in
            self.selected_sub = self.nsubs - 1
        else:
            self.update_stack_scroller()

    def delete_selected_sub(self):
        if self.nsubs == 0:
            return

        try:
            self.scope.stacker.delete_selected_sub(self.selected_sub)
        except Exception as e:
            # will have been dealt with re notification by this point
            return

        # sub has been removed from stack and shifted to delete subfolder
        self.nsubs -= 1

        if self.nsubs == 0:
            # nothing, so act like a clear_stack operation
            self.scope.aligner.reset()
            self.scope.stacker.reset()
            self.reset()

        elif self.selected_sub == self.nsubs:
            # was at end of stack, so shift down one
            self.selected_sub -= 1

        if self.selected_sub < self.nsubs:
            # fake a call to on_selected_sub
            self.on_selected_sub()
        else:
            # deleted last one, so shift selected_sub back one; this will
            # call on_selected_sub and update display
            self.selected_sub -= 1

    def on_stack_combine_method(self, *args):
        if self.sub_or_stack == 'stack':
            self.on_selected_sub()

    def on_selected_sub(self, *args):
        # key method: when a sub is selected, eventually display gets updated
        # This should be the only route to display updates.

        if not hasattr(self.scope, 'stacker'):
            return

        s = self.selected_sub
        if (s < 0) or (s >= self.nsubs):
            return

        if self.sub_or_stack == 'sub':
            sub = self.scope.stacker.subs[s]
            self.image = sub.image
            self.scope.stacker.selected = []

        else:
            new_image = self.scope.stacker.compute_stack(up_to=s)
            if new_image is not None:
                self.image = new_image

        self.update_image()
        self.update_stack_scroller()

    # def on_stack_position(self, *args):
    #     self.selected_sub = min(self.nsubs - 1, max(0, round(self.nsubs * self.stack_position) - 1))

    def toggle_sub_or_stack(self, *args):
        newval = 'stack' if self.sub_or_stack == 'sub' else 'sub'
        self.sub_or_stack = newval
        self.stack_toggle.text = newval
        self.on_selected_sub()

    def on_zoom(self, *args):
        # zoom anchored on window centre
        if self.scatter is not None:
            scale = self.zoom / self.scatter._get_scale()
            mat = Matrix().scale(scale, scale, scale)
            self.scatter.apply_transform(mat, anchor=self.origin)
            #print('   rot:', self.scatter._get_rotation())
            # print('centre: {:} scale: {:}'.format(self.scatter._get_center(), self.scatter._get_scale()))
            #print(' local:', self.scatter.to_local(*self.scatter._get_center()))

    def on_orientation(self, *args):
        # rotate anchored on window center
        if self.scatter is not None:
            rot = self.orientation - self.scatter._get_rotation()
            self.scatter.apply_transform(Matrix().rotate((np.pi/180)*rot, 0 ,0, 1), anchor=self.origin)

    def on_reset(self, *args):
        self.levers['zoom'].reset_value(1)
        self.zoom = 1
        self.orientation = 0
        self.scatter._set_center(self.origin)

    def on_flip_LR(self, *args):    self.display_image()
    def on_flip_UD(self, *args):    self.display_image()
    def on_gradient(self, *args):   self.display_image()
    def on_black(self, *args):      self.display_image()
    def on_white(self, *args):      self.display_image()
    def on_p1(self, *args):         self.display_image()
    def on_stretch(self, *args):    self.display_image()
    def on_greylevel(self, *args):  self.display_image()

    # def on_binning(self, *args):
    #     self.binning_on = self.binning > 1.1
    #     self.scope.stacker.check_recompute()
    #     self.display_image()

    # we sandwich between reticule toggling otherwise nothing to update on the screen
    def on_invert(self, *args):

        if not hasattr(self, 'levers'):
            return

        if hasattr(self, 'reticule'):
            self.on_reticule()
            self.display_image()
            self.on_reticule()

        if not self.hypernegative:
            return

        if self.invert:
            self.invert_settings = {'stretch': self.stretch, 'white': self.white, 'p1': self.p1 }
            self.stretch = 'hyper'
            self.levers['white'].reset_value(.85)
            self.levers['p1'].reset_value(.9)
        elif hasattr(self, 'invert_settings'):
            self.stretch = self.invert_settings['stretch']
            self.levers['white'].reset_value(self.invert_settings['white'])
            self.levers['p1'].reset_value(self.invert_settings['p1'])

    def on_autoblack(self, *args):
        if self.autoblack > 1:
            ab = 15 - self.autoblack
            if hasattr(self, 'mean_g') and self.mean_g:
                #Logger.debug('Eyepiece: setting autoblack to {:8.6f}'.format(self.mean_g - ab*self.std_g))
                self.levers['black'].reset_value(self.mean_g - ab*self.std_g)

    def on_ring(self, *args):
        self.scope.app.ring_transparency = min(1, max(0, self.ring))

    def on_fine(self, *args):
        self.black = min(1, max(0, self.black + self.fine * self.fine_sensitivity))
        self.levers['black'].reset_value(self.black)
 
    # experimental
    # def on_ROI(self, *args):
    #     self.apply_ROI = not self.apply_ROI
    #     if self.apply_ROI & (self.image is not None):
    #         h, w = self.image.shape
    #         # print(w, h)
    #         ix, iy = w/2, h/2                    # image center in pixel coords
    #         sx, sy = self.scatter._get_center()  # scatter center in screen coords
    #         ex, ey = self.origin                 # eyepiece origin in screen coords
    #         x = ix + ((ex - sx) / self.zoom)       # image center under eyepiece in pixel coords
    #         y = iy + ((ey - sy) / self.zoom)
    #         r = self.inner_radius / self.zoom          # image radius in pixel coords
    #         self.ROI = [max(int(x - r), 0), max(int(y - r), 0), min(int(x + r), w), min(int(y + r), h)]

    #         # to do: reverse coords if fliplr and ud
    #         # not yet working
    #         if self.flip_UD:
    #             top, bottom = h - self.ROI[3], h - self.ROI[1]
    #             self.ROI[1] = top
    #             self.ROI[3] = bottom

    #         if self.flip_LR:
    #             left, right = w - self.ROI[0], w - self.ROI[2]
    #             self.ROI[0] = right
    #             self.ROI[2] = left
                
    #         print('image center (px) {:}'.format((ix, iy)))
    #         print('  eye center (px) {:}'.format((x, y)))
    #         print('      radius (px) {:}'.format(r))
    #         print(' scat center (sc) {:}'.format((sx, sy)))
    #         print('  eye center (sc) {:}'.format((ex, ey)))
    #         print('      radius (sc) {:}'.format(self.inner_radius))
    #         print('ROI {:}'.format(self.ROI))
    #     else:
    #         self.ROI = None

    def on_reticule(self, *args):
        self.show_reticule = not self.show_reticule
        #self.reticule_button.color = [1, 0, 0, 1] if self.show_reticule else [.5, .5, .5, 1]
        self.reticule.show = self.show_reticule

    def on_speed(self, *args):
        if hasattr(self, 'playing'):
            if self.playing:
                Clock.unschedule(self.play_event)
                self.play_event = Clock.schedule_interval(self.increment_sub, 1/self.speed)

    # called whenever the image itself changes as opposed to when the existing image
    # brightness mapping is manipulated etc (handled in display_image below)
    def update_image(self):

        if self.redrawing or (self.image is None):
            return

        self.just_updated = False

        # recompute grids if shape changes
        if (self.scatter.w, self.scatter.h) != self.image.shape:
            w, h = self.image.shape
            self.grid_x, self.grid_y = np.meshgrid(np.arange(h), np.arange(w))

        # compute background gradient for new image
        # in some situations this can fail
        grad = estimate_gradient(self.image, self.grid_x, self.grid_y)
        if grad: 
            self.background_gradient, self.mean_g, self.std_g = grad
            self.std_g = np.std(self.image) / 100
        else:
            self.background_gradient = None
            self.mean_g = None
            self.scope.stacker.subs[self.selected_sub].describe()

        if self.autoblack > 1:
            self.on_autoblack()
        
        if not self.just_updated:
            self.display_image()

    # called when image needs to be redisplayed (control changed, new image)
    def display_image(self, dt=0):

        # relative timings:
        #  gradient, rescale and stretch each take around 30% of the time
        #  image display takes the rest; invert, flipUD, flipLR are negligible
        #  different stretches take more or less time; hyper is fast, arcsinh is slow
        # not really worth caching intermediate results yet; takes around 70ms for a 1600x1200 mono image
        # but when handling multispectral probably worth operating directly on the blit buffer representation?

        if self.redrawing or (not hasattr(self, 'image')) or (self.image is None):
            return

        x = self.image

        if self.gradient > 0:
            if self.background_gradient is not None:
                x = x - self.gradient/100 * self.background_gradient

        # preview binning if current stack has bin = 1
        # if self.binning_on:
        #     if self.scope.stacker.binning() < 1.1:
        #         x = do_binning(x, self.binning, original_size=True)

        w, b  = self.white, self.black

        # xs is input to stretch
        xs = (x - b) / (w - b)
        xs[xs > 1] = 1
        xs[xs < 0] = 0

        # result of stretch in range 0-1
        xs_out = stretch(xs, method=self.stretch, param=self.p1)
        x = xs_out

        # shift it to range grey - 1
        #xs_out = (1 - self.greylevel) * xs_out + self.greylevel

        # insert stretched where originally above background
        #x[x > b] = xs_out[x > b]

        # and linear gradient below that
        # x[x <= b] = self.greylevel  # x[x <= b] * (self.greylevel / b)
        #x[x <= b] = x[x <= b] * (self.greylevel / b)

        # for now
        # x = xs_out

        if self.invert:
            x = np.subtract(1, x, dtype=x.dtype)

        if self.flip_UD: 
            x = x[::-1]

        if self.flip_LR:
            x = np.fliplr(x)

        self.scatter.display(x)

        self.just_updated = True  # to prevent multiple updates

