''' Notifier. Lightweight class to show transient notifications. Can be switched off by user.
'''

from kivy.uix.boxlayout import BoxLayout
from kivy.clock import Clock
from kivy.animation import Animation
from kivy.core.window import Window
from kivy.properties import BooleanProperty
from kivy.lang import Builder

from jocular.utils.component import Component

Builder.load_string('''

<Notifier>:
    orientation: 'horizontal'
    size_hint: None, None
    width: dp(220)
    height: dp(52)
    margin: 1
    state: 'hidden'
    contents: _contents

    canvas.before:
        Color:
            rgba: .15, .2, .15, 1
        Rectangle:
            pos: self.pos
            size: self.size

    TextInput:
        id: _contents
        background_color: (.2, .2, .2, 1)
        foreground_color: (1, 1, 1, app.dimmed_transp)
        markup: True
        text_size: self.size
        multiline: True
        size_hint_x: 1
        margin: dp(4)
        font_size: '14sp'
''')

class Notifier(BoxLayout, Component):

    show_notifications = BooleanProperty(True)

    settables = ['show_notifications']

    def __init__(self, scope):

        self.scope = scope
        super().__init__()
        self.notification = ''
        self.load_settings(settings=scope.settings)

    def redraw(self, *args):
        self.scope.add_widget(self, index=0)
        try:
            self.anim.cancel(self)
            self.denotify_event.cancel()
        except AttributeError:
            pass
        finally:
            self.state = 'hidden'
            self.denotify_event = Clock.schedule_once(self.denotify, -1)
            self.pos = Window.width, self.margin

    def notify(self, message, duration=5):

        if not self.show_notifications:
            return

        self.contents.text = message

        # normal case. Nothing already shown
        if self.state == 'hidden':
            self.anim = Animation(x=Window.width - self.width - self.margin, duration=0.5)
            self.anim.bind(on_start=self.notifier_incoming, on_complete=self.notifier_displayed)
            self.anim.start(self)
            self.denotify_event = Clock.schedule_once(self.denotify, duration)  # before next frame

        elif self.state == 'incoming':
            pass
 
        elif self.state == 'displayed':
            # cancel denotify and reschedule
            self.denotify_event.cancel()
            self.denotify_event = Clock.schedule_once(self.denotify, 0)  # before next frame

        elif self.state == 'leaving':
            # cancel animation and set up new one of shorter duration
            self.anim.cancel(self)
            self.denotify_event.cancel()
            self.anim = Animation(x=Window.width - self.width - self.margin, duration=0.1)
            self.anim.bind(on_start=self.notifier_incoming, on_complete=self.notifier_displayed)
            self.anim.start(self)
            self.denotify_event = Clock.schedule_once(self.denotify, 0)  # before next frame

    def denotify(self, *args):
        self.anim = Animation(x=Window.width, duration=1)
        self.anim.bind(on_start=self.notifier_leaving, on_complete=self.notifier_hidden)
        self.anim.start(self)

    def notifier_displayed(self, *args):
        self.state = 'displayed'

    def notifier_incoming(self, *args):
        self.state = 'incoming'

    def notifier_leaving(self, *args):
        self.state = 'leaving'

    def notifier_hidden(self, *args):
        self.contents.text = ''
        self.state = 'hidden'
