''' Obj. Represents the current object and builds the object panel. Loads/saves metadata.
'''

import os
import json
import operator
import shutil
import contextlib
from datetime import datetime, date

from kivy.logger import Logger
from kivy.event import EventDispatcher
from kivy.properties import NumericProperty, StringProperty, OptionProperty, ListProperty, BooleanProperty, ObjectProperty
from kivy.clock import Clock
from kivy.lang import Builder
from kivy.core.window import Window
from kivy.uix.boxlayout import BoxLayout
from kivy.animation import Animation
from kivy.metrics import dp

from jocular.utils.component import Component
from jocular.image.sub import Sub
from jocular.utils.utils import unique_dir_name, add_if_not_exists, datenow, \
    remove_nulls_from_dict, get_fits, date_time_format, hours_since_date, \
    dodgy_name, purify_name, joc_icon
from jocular.widgets.widgets import ConfigToggle

Builder.load_string('''

<MyTextInput@TextInput>:
    background_color: app.background_color
    foreground_color: app.text_color
    background_normal: ''
    multiline: False
    hint_text_color: app.lowlight_color
    font_size: '16sp'
    border: dp(4), dp(4), dp(4), dp(4)

<Obj>:
    canvas.before:
        Color:
            rgba: app.line_color
        Rectangle:
            pos: self.pos
            size: self.size
    canvas.after:
        Color:
            rgba: app.line_color
        Line:
            points: self.x + self.width, self.y, self.x + self.width, self.y + self.height
            width: .5

    size_hint: None, None
    orientation: 'vertical'
    spacing: 1

    BoxLayout:
        orientation: 'horizontal'
        size_hint: 1, None
        height: dp(40)
        spacing: 1
        MyTextInput:
            id: _name
            size_hint: 1, 1
            font_size: '24sp'
            valign: 'middle'
            text: root.Name
            hint_text: 'Object name:'
            on_text: root.Name = self.text
            on_text_validate: root.set_name(self.text)


    BoxLayout:
        orientation: 'horizontal'
        size_hint: 1, None
        height: dp(32)
        spacing: 1
        MyTextInput:
            size_hint: 0.12, 1
            text: root.Con
            hint_text: 'con'
            on_text: root.Con = self.text
        MyTextInput:
            size_hint: 0.08, 1
            hint_text: 'typ'
            text: root.OT
            on_text: root.OT = self.text
        MyTextInput:
            size_hint: 0.25, 1
            hint_text: 'ra'
            text: str(root.RA)
            on_text: root.RA = self.text
        MyTextInput:
            size_hint: 0.25, 1
            hint_text: 'Dec'
            text: str(root.Dec)
            on_text: root.Dec = self.text
        MyTextInput:
            size_hint: .16, 1
            hint_text: 'mag'
            text: 'm{:}'.format(root.Mag) if root.Mag else ''
            on_text: root.Mag = self.text[1:]
        MyTextInput:
            size_hint: .14, 1
            hint_text: 'size'
            text: "{:}'".format(root.Diam) if root.Diam else ''
            on_text: root.Diam = self.text[:-1]

    BoxLayout:
        orientation: 'horizontal'
        size_hint: 1, None
        height: dp(32)
        spacing: 1
        MyTextInput:
            size_hint: .2, 1
            hint_text: 'distance'
            text: root.Dist
            on_text: root.Dist = self.text
        MyTextInput:
            size_hint: .8, 1
            hint_text: 'also known as'
            text: root.AKA
            on_text: root.AKA = self.text

    BoxLayout:
        orientation: 'horizontal'
        size_hint: 1, None
        height: dp(32)
        spacing: 1
        MyTextInput:
            size_hint: 1, 1
            hint_text: 'other'
            text: root.Other
            on_text: root.Other = self.text

    MyTextInput:
        size_hint: 1, 1
        multiline: True
        on_text: root.Notes = self.text
        text: root.Notes
        hint_text: 'Observing notes:'

    MyTextInput:
        text: 'Observing conditions {:}'.format(root.session)
        size_hint: 1, None
        height: dp(32)

    BoxLayout:
        canvas.before:
            Color:
                rgba: app.background_color
            Rectangle:
                pos: self.pos
                size: self.size

        spacing: 1
        orientation: 'horizontal'
        size_hint: 1, None
        height: dp(100)
        Label:
            size_hint: .1, 1
        BoxLayout:
            size_hint: None, None
            size: dp(100), dp(100)
            JKnob:
                knob_type: 'numeric'
                min: 16
                max: 22
                value_format:       '{:5.2f}'
                value: root.SQM
                title: 'SQM'
                presets: []
                on_value: root.SQM = self.value
                inactive: root.SQM_inactive
                on_inactive: root.SQM_inactive = self.inactive
        Label:
            size_hint: .1, 1
        BoxLayout:
            size_hint: None, None
            size: dp(100), dp(100)
            JKnob:
                knob_type: 'hidden presets'
                value: root.transparency
                presets: ['not set', 'cloudy', 'v hazy' ,'cirrus', 'light haze', 'clear', 'v clear', 'excellent']
                on_value: root.transparency = self.value
                title: 'transparency'
        Label:
            size_hint: .1, 1
        BoxLayout:
            size_hint: None, None
            size: dp(100), dp(100)
            JKnob:
                knob_type: 'hidden presets'
                size: dp(100), dp(100)
                value: root.seeing
                presets: ['not set', 'disturbed', 'poor' ,'good', 'excellent', 'superb']
                on_value: root.seeing = self.value
                title: 'seeing'
        Label:
            size_hint: .1, 1

    MyTextInput:
        size_hint: 1, None
        height: dp(116)
        multiline: True
        on_text: root.session_notes = self.text
        text: root.session_notes
        hint_text: 'Session notes:'
        disabled: False

    BoxLayout:
        orientation: 'horizontal'
        size_hint: 1, None
        height: dp(32)
        spacing: 1
        MyTextInput:
            size_hint: .5, 1
            text: root.telescope
            hint_text: 'scope:'
            on_text: root.telescope = self.text
            disabled: False
        MyTextInput:
            size_hint: .5, 1
            on_text: root.camera = self.text
            hint_text: 'camera:'
            text: root.camera
            disabled: False

    MyTextInput:
        size_hint: 1, None
        height: dp(32)
        on_text: root.other_kit = self.text
        hint_text: 'other kit:'
        text: root.other_kit
        disabled: False

    MyTextInput:
        size_hint: 1, None
        height: dp(32)
        hint_text: 'Observer'
        text: root.observer
        on_text: root.observer = self.text
''')

class Obj(BoxLayout, EventDispatcher, Component):

    # DSO information
    Name = StringProperty('')
    Con = StringProperty('')
    RA = StringProperty('')
    Dec = StringProperty('')
    Dist = StringProperty('')
    Diam = StringProperty('')
    OT = StringProperty('')
    Mag = StringProperty('')
    Other = StringProperty('')
    AKA = StringProperty('')
    Notes = StringProperty('')

    # session/observing conditions
    session = StringProperty('')
    SQM = NumericProperty(17.0)
    temperature = NumericProperty(5)
    seeing = StringProperty('not set')
    transparency = StringProperty('not set')
    session_notes = StringProperty('')
    SQM_inactive = BooleanProperty(True)  # in case user hasn't set these values
    temperature_inactive = BooleanProperty(True)
    date_of_last_session = StringProperty(None)
    retain_session_hours = NumericProperty(3)

    # kit and observer info
    observer = StringProperty('')
    telescope = StringProperty('')
    camera = StringProperty('')
    other_kit = StringProperty('')

    # capture information
    exposure = NumericProperty(1)
    sub_type = OptionProperty('light', options=['light', 'dark', 'flat', 'bias', 'RGB', 'LRGB', 'short'])
    selected = ListProperty([])
    rejected = ListProperty([])
    nonaligned = ListProperty([])
    keyframe = StringProperty('')

    # calibration masters
    original_masters = ObjectProperty({'dark': None, 'flat': None, 'bias': None})  # those USED originally
 
    # master name (only if this obj instance is a calibration object)
    master_name = StringProperty('')

    # list of properties to enable cleaner code when loading etc
    obj_props = ['Name', 'Con', 'RA', 'Dec', 'Dist', 'Diam', 'OT', 'Mag', 'Other', 'AKA', 'Notes']
    session_props = ['session', 'SQM', 'temperature', 'seeing', 'transparency', 'session_notes']
    kit_props = ['telescope', 'camera', 'other_kit', 'observer']
    capture_props = ['exposure', 'sub_type', 'rejected', 'nonaligned', 'keyframe', 'original_masters', 'master_name']

    auto_save = BooleanProperty(False)
 
    # show/hide
    state = OptionProperty('displayed', options=['displayed', 'leaving', 'hidden', 'incoming'])
    pinned = BooleanProperty(False)

    # properties that are saved in between sessions
    settables = [
        'observer', 'SQM', 'temperature', 'seeing', 'transparency',
        'telescope', 'camera', 'other_kit', 'session_notes', 'pinned', 'auto_save',
        'temperature_inactive', 'SQM_inactive', 'date_of_last_session', 'retain_session_hours'
        ]

    def __init__(self, scope, **kwargs):

        Logger.debug('Obj: initialising')
        self.scope = scope
        super().__init__(**kwargs)
        self.load_settings(settings=scope.settings)

        # set session props by saving (to create live_session_properties), then restoring
        self.save_session_settings()
        self.restore_session_settings()

        self.width = dp(420)
        self.pos = -self.width, 0

    def post_init(self, *args):
        self.clear_object_info()
        self.clear_capture_props()

    def pin_pressed(self, inst, value):
        self.pinned = not self.pinned

    def redraw(self, *args):

        Logger.debug('Obj: redrawing')

        self.scope.add_widget(self)
        self.size = dp(420), Window.height

        # add pin just outside box
        self.pin = ConfigToggle(text='', size_hint=(None, None), size=(28, 28))
        self.pin.bind(state=self.pin_pressed)
        self.pin.y = Window.height - self.pin.height
        self.scope.add_widget(self.pin)
 
        if self.pinned:
            self.pin.x = self.width
            self.pin.text = joc_icon('pinned') # '[font=Jocular][size=16]1[/size][/font]'

            if self.scope.processing.pinned:
                self.pos = 0, self.scope.processing.height
                self.height = Window.height - self.scope.processing.height
            else:
                self.pos = 0, 0
        else:
            self.pin.x = -self.pin.width
            self.pos = -self.width, 0
            self.pin.text = joc_icon('unpinned') # '[font=Jocular][size=16]2[/size][/font]'

    def on_close(self, *args):
        # save any unsaved object metadata
        self.save()
        # save last live session settings; if already live, no need to restore
        if not self.scope.live:
            self.restore_session_settings()
        if hasattr(self, 'saved_settings'):
            self.restore_settings()

    # ensure no clicks get passed thru
    def on_touch_down(self, touch):
        handled = super().on_touch_down(touch)
        if self.collide_point(*touch.pos):
            return True
        return handled

    def on_temperature(self, *args):
        if self.scope.closing:
            return
        if hasattr(self.scope, 'calibrator'):
            self.scope.calibrator.update_masters()

    def clear_object_info(self, clear_name=True, clear_notes=True):
        for p in self.obj_props:
            if (p == 'Name') and not clear_name:
                pass
            elif (p == 'Notes') and not clear_notes:
                pass
            else:
                setattr(self, p, '')

    def clear_capture_props(self):
        for p in self.capture_props:
            if p not in ['sub_type', 'exposure']:
                setattr(self, p, '')

    def clear_stack(self, *args):
        # test version

        if not self.scope.live:
            self.scope.notifier.notify('Only allowed in live mode')
            return

        # delete existing subs
        try:
            # delete all subs in self.save_dir
            for nm in get_fits(self.save_dir):
                os.remove(nm)
        except Exception as e:
            Logger.exception('Obj: problem removing files when clearing stack')

        self.scope.aligner.reset()
        self.scope.stacker.reset()
        self.scope.eyepiece.reset()

    def new_from_menu(self, *args):
        # needed to soak up args when calling from widget
        Logger.debug('Obj: new object selected by user')
        self.new()

    def new(self, dso=None, stop_camera=True, save=True):
        '''set up for a new object; if called with dso not None, fill in dso details
        '''

        Logger.debug('Obj: new object')

        if stop_camera:
            self.scope.camera.stop()

        if save:
            self.save()

        # if we were offline, restore most recent live settings
        if not self.scope.live:
            self.restore_session_settings()

        # update live session date
        self.date_of_last_session = datenow()

        self.scope.aligner.reset()
        self.scope.stacker.reset()
        self.scope.eyepiece.reset()

        self.scope.live = True
        self.scope.watcher.start()

        # create session directory if it doesn't exist
        sesh_dir = os.path.join(self.scope.captures_dir, date.today().strftime('%d_%b_%y'))
        add_if_not_exists(sesh_dir)

        # clear object and capture panels
        self.clear_object_info()
        self.clear_capture_props()

        # create a new folder for this object
        if dso is None:
            self.Name = ''
        else:
            self.set_name(dso)

        self.save_dir = os.path.join(sesh_dir, unique_dir_name(self.Name, sesh_dir))

        # set up session seconds counter
        self.session_clock = Clock.schedule_interval(self.update_time, 1.0)

    def update_time(self, *args):
        self.session = datetime.now().strftime(date_time_format)

#===============================
# session handling
#   eventually move to own class
#===============================

    def save_session_settings(self):
        Logger.debug('Obj: saving live session settings')
        self.live_session_properties = {
            'SQM': self.SQM,
            'temperature': self.temperature,
            'seeing': self.seeing,
            'transparency': self.transparency,
            'session_notes': self.session_notes,
            'SQM_inactive': self.SQM_inactive,
            'temperature_inactive': self.temperature_inactive
        }

    def restore_session_settings(self):
        if self.date_of_last_session is not None:
            if hours_since_date(self.date_of_last_session) < self.retain_session_hours:
                # restore
                Logger.debug('Obj: restoring last live session values')
                for k, v in self.live_session_properties.items():
                    setattr(self, k, v)
            else:
                # set to sensible initial values
                Logger.debug('Obj: resetting live session values (out of date)')
                self.session_notes = ''
                self.SQM_inactive = True
                self.temperature_inactive = True
                self.seeing = 'not set'
                self.transparency = 'not set'
                self.SQM = 17
                self.temperature = 5

#===========================

    def delete(self, *args):
        # Mark object for deletion and set up for new object

        Logger.debug('Obj: delete: marking for deletion')
        if not self.Name.startswith('DELETE '):
            self.Name = 'DELETE {:}'.format(self.Name)
        self.new()

#============================
# loading and saving metadata
#============================

    def load(self, path):
        '''load existing object'''

        Logger.debug('Obj: load object from {:}'.format(path))

        if (not path) or (not os.path.exists(os.path.join(path, 'info.json'))):
            Logger.info('Obj: load: no info.json')
            return

        if hasattr(self, 'save_dir'):
            if path == self.save_dir:
                # object already loaded, so do a new followed by load
                self.new(save=False)
                self.load(path)
                return

        # save last object with its settings
        self.save()

        # if we were in live mode, save session info
        if self.scope.live:
            self.save_session_settings()

        self.scope.aligner.reset()
        self.scope.stacker.reset()
        self.scope.eyepiece.reset()
        self.scope.live = False
        self.scope.watcher.stop()
        self.scope.camera.stop()
        self.estimating_exposure = False  # not used on load

        # halt session clock
        if hasattr(self, 'session_clock'):
            Clock.unschedule(self.session_clock, -1)

        self.save_dir = path

        # load metadata, returning if any problem
        try:
            with open(os.path.join(path, 'info.json'), 'r') as f:
                hdr = json.load(f)
        except json.JSONDecodeError as e:
            self.scope.notifier.notification('Failed to load. Problem reading metadata')
            return

        # ensure original session is cached
        self.loaded_session = hdr['session']

        # clear and apply metadata where it exists
        self.clear_capture_props()
        self.clear_object_info()

        # inactivate knobs if values not set
        self.SQM_inactive = 'SQM' not in hdr
        self.temperature_inactive = 'temperature' not in hdr

        if 'seeing' not in hdr: self.seeing = 'not set'
        if 'transparency' not in hdr: self.transparency = 'not set'
        self.session_notes = '' 

        # Apply header information to update levers etc
        for k, v in hdr.items():
            if k == 'scope': # updating scope properties
                for k1, v1 in v.items():
                    if hasattr(self.scope, k1):
                        setattr(self.scope, k1, v1)
                        if k1 in self.scope.eyepiece.levers:
                            self.scope.eyepiece.levers[k1].reset_value(v1)
            elif k != 'image_center':
                if hasattr(self, k):
                    setattr(self, k, v)

        # set image centre to last saved position
        # this has to be done last of all (after setting zoom/orientation)
        if 'image_center' in hdr:
            cntr = hdr['image_center']
            self.scope.eyepiece.scatter._set_center((cntr[0], cntr[1]))

        # faster loading if we don't do stack combination during the load
        if self.scope.eyepiece.sub_or_stack == 'stack':
            self.scope.eyepiece.toggle_sub_or_stack(None)

        # set up names to load in correct order (this has support for future realign)
        if 'rejected' not in hdr:
            hdr['rejected'] = []
        if 'nonaligned' not in hdr:
            hdr['nonaligned'] = []

        # get FITs in data order; any problems, just concatenate
        all_fits = get_fits(path)

        if len(all_fits) == 0:
            Logger.info('Obj: no FITs in {:}'.format(path))
            self.scope.notifier.notify('no FITs')
            return

        try:
            names = {os.path.basename(nm): os.path.getmtime(nm) for nm in all_fits} 
            names = [n for n, _ in sorted(names.items(), key=operator.itemgetter(1))]
        except OSError as e:
            Logger.debug('Obj: cannot get modtime for FITs. Choosing arbitrary order.')
            names = [os.path.basename(nm) for nm in all_fits]

        if 'keyframe' not in hdr:
            hdr['keyframe'] = names[0]

        # schedule load
        self.names, self.align_directions = self.scope.stacker.reorder_subs_sequence(names, hdr['keyframe'])
        self.hdr = hdr
        self.count = 0
        self.abort_load = False
        Clock.schedule_once(self._load_subs, 0)

    def escape_pressed(self):
        self.abort_load = True

    def _load_subs(self, dt):
        if self.count < len(self.names):
            nm, align_direction = self.names[self.count], self.align_directions[self.count]
            self.add_sub(nm, 
                sub_type=self.hdr['sub_type'], 
                rejected=nm in self.hdr['rejected'], 
                align_direction=align_direction)
            self.count += 1
            if not self.abort_load:
                Clock.schedule_once(self._load_subs, 0.3)
            else:
                self.new(save=False)

    def save(self):
        # Saves current object metadata, If object is a calibration stack, saves master.

        if len(self.scope.stacker.subs) == 0:
            return

        Logger.info('Obj: saving')
        self.Name = self.Name.strip()

        delete = self.Name.startswith('DELETE ')

        calibration = self.sub_type in {'dark', 'flat', 'bias'}

        # update sub status
        self.rejected = [s.name for s in self.scope.stacker.subs if s.status == 'reject']
        self.nonaligned = [s.name for s in self.scope.stacker.subs if s.status == 'nalign']

        if calibration:
            # if calibration sub, save master also and store master name
            # Dont save calibration master if delete is requested
            if not delete:
                calib_master_name = self.scope.calibrator.save_master(master_name=self.master_name)
                if calib_master_name:
                    self.master_name = calib_master_name
                    if len(self.Name.strip()) == 0: 
                        self.Name = calib_master_name  # use this name if user has not provided one

        # fits saved in self.save_dir, but at this point we might have a new name
        orig_save_dir = self.save_dir
        session_dir, orig_name = os.path.split(orig_save_dir)

        # issue a warning if string contains non-alphanumeric values
        if dodgy_name(self.Name):
            self.Name = purify_name(self.Name)
            Logger.debug('Obj: purified name to {:}'.format(self.Name))
            self.scope.notifier.notify('Forbidden characters in name removed')

        # check if orig_name differs from user provided name (but only if not marked for deletion)
        if (orig_name != self.Name) and (not delete):
            # generate a unique folder name and rename old to new
            try:
                unique_name = unique_dir_name(self.Name, session_dir)
                # self.Name = unique_name
                self.save_dir = os.path.join(session_dir, unique_name)
                Logger.debug('Obj: changing save dir from {:} to {:}'.format(orig_save_dir, self.save_dir))
                os.rename(orig_save_dir, self.save_dir)
                self.scope.observations.rename(old_path=orig_save_dir, new_path=self.save_dir)
            except OSError:
                # might be an odd character in name, for instance
                self.scope.notifier.notify('Unable to rename as {:}'.format(self.Name))
                # so restore original name
                self.save_dir = orig_save_dir

        # if we are live, store masters
        if self.scope.live:
            self.original_masters = self.scope.calibrator.masters

        # build metadata from object properties
        hdr = {}
        for p in self.obj_props + self.session_props + self.kit_props + self.capture_props:
            hdr[p] = getattr(self, p)

        # overwrite session with original if not live
        if not self.scope.live:
            hdr['session'] = self.loaded_session

        # modify the inactive controls
        if self.SQM_inactive: del hdr['SQM']
        if self.temperature_inactive: del hdr['temperature']

        # save scope parameters
        eyepiece = self.scope.eyepiece
        hdr['scope'] = {'zoom': eyepiece.zoom, 'orientation': eyepiece.orientation, 'black': eyepiece.black, 
            'white': eyepiece.white, 'p1': eyepiece.p1, 'gradient': eyepiece.gradient, 'invert': eyepiece.invert, 
            'stretch': eyepiece.stretch, 'stack_combine_method': eyepiece.stack_combine_method}

        # store image center so it can be restored
        if hasattr(self.scope.eyepiece, 'scatter'):
            if self.scope.eyepiece.scatter:
                hdr['image_center'] = self.scope.eyepiece.scatter._get_center()

        hdr['n_subs'] = len(self.scope.stacker.subs)

        # remove any null fields
        hdr = remove_nulls_from_dict(hdr)

        # save metadata
        try:
            with open(os.path.join(self.save_dir, 'info.json'), 'w') as f:
                json.dump(hdr, f, indent=4)
            if self.scope.live and (not delete):
                self.scope.notifier.notify('Saved to {:}'.format(os.path.basename(self.save_dir)))
        except json.JSONDecodeError as e:
            Logger.debug('Obj: problem saving info.json to {:}'.format(self.save_dir))
            self.scope.notifier.notify('Metadata not saved')

        # update observations
        self.scope.observations.update(self.save_dir, hdr)

        if calibration:
            self.restore_settings()
        elif (not delete) and self.auto_save and self.scope.live:
            try:
                self.scope.snapshotter.save()
            except Exception as e:
                Logger.exception('Obj: problem saving snapshot')

#===================

    def apply_calibration_settings(self, sub_type=None):
        # save current settings and apply calibration-specific settings
        
        Logger.debug('Obj: applying calibration-specific settings')

        eyepiece = self.scope.eyepiece

        self.saved_settings = {
            'black': eyepiece.black,
            'white': eyepiece.white,
            'gradient': eyepiece.gradient,
            'autoblack': eyepiece.autoblack,
            'p1': eyepiece.p1,
            'stack_combine_method': eyepiece.stack_combine_method,
            'stretch': eyepiece.stretch,
            'apply_flat': self.scope.calibrator.apply_flat}

        # create suitable settings for calibration frames
        eyepiece.levers['gradient'].reset_value(0)
        eyepiece.levers['autoblack'].reset_value(7)

        if sub_type == 'flat':
            eyepiece.stack_combine_method = '80%'
            eyepiece.levers['black'].reset_value(.4)
            eyepiece.levers['white'].reset_value(.6)
            eyepiece.stretch = 'linear'
            self.scope.calibrator.apply_flat = True  # ensures bias is loaded if available
        elif sub_type == 'dark':
            eyepiece.stack_combine_method = 'mean'
            eyepiece.stretch = 'hyper'
            eyepiece.levers['p1'].reset_value(.9)
        elif sub_type == 'bias':
            eyepiece.stack_combine_method = 'mean'
            eyepiece.stretch = 'hyper'
            eyepiece.levers['p1'].reset_value(.9)
            eyepiece.levers['white'].reset_value(.2)

    def restore_settings(self):

        Logger.debug('Obj: restoring light-specific settings')

        s = self.saved_settings
        eyepiece = self.scope.eyepiece
        eyepiece.levers['black'].reset_value(s['black'])
        eyepiece.levers['white'].reset_value(s['white'])
        eyepiece.levers['gradient'].reset_value(s['gradient'])
        eyepiece.levers['autoblack'].reset_value(s['autoblack'])
        eyepiece.levers['p1'].reset_value(s['p1'])
        eyepiece.stretch = s['stretch']
        eyepiece.stack_combine_method = s['stack_combine_method']
        self.scope.calibrator.apply_flat = s['apply_flat']

    def on_exposure(self, *args):

        Logger.debug('Obj: exposure changed to {:5.2f}s'.format(self.exposure))

        if hasattr(self, 'estimating_exposure') and hasattr(self, 'auto_exposure'):
            if not self.auto_exposure:
                self.estimating_exposure = False

        if hasattr(self.scope, 'calibrator'):
            self.scope.calibrator.update_masters()

    def add_sub(self, name, from_dir=None, sub_type=None, exposure=None, align_direction='forward', rejected=False):
        # Creates new sub and adds to stack. Called by Watcher and from Observations.load

        Logger.debug('Obj: add sub of type {:}'.format(sub_type))

        try:
            first_sub = len(self.scope.stacker.subs) == 0

            # check for subtype change only if not first sub
            if not first_sub:
                if sub_type != self.sub_type:
                    Logger.debug('Obj: sub type changed to {:}'.format(sub_type))
                    self.scope.notifier.notify('Sub type changed to {:}. Saving.'.format(sub_type))
                    self.new(stop_camera=False)
                    first_sub = True

            # if it is first sub, set sub type and do any necessary initialisations
            if first_sub:
                self.sub_type = sub_type
                if sub_type in ['dark', 'flat', 'bias']:
                    self.apply_calibration_settings(sub_type)
                if self.scope.live:
                    self.estimating_exposure = True
                    self.scope.watcher.reset_exposure_estimate()

            # move sub to save dir (we do this after checking for new sub type since save dir might have changed)
            # we only supply a from_dir for monitored captures
            if from_dir is not None:
                add_if_not_exists(self.save_dir)
                if os.path.exists(os.path.join(self.save_dir, name)):
                    with contextlib.suppress(FileNotFoundError):
                        os.remove(os.path.join(from_dir, name))
                else:
                    try:
                        shutil.move(os.path.join(from_dir, name),  self.save_dir)
                    except OSError as e:
                        Logger.exception('Obj: add_sub')

            # handle exposure
            # if live mode, estimating_exposure is set to True on encountering the first sub
            # if at any subsequent point the user provides a new exposure, estimating exposure
            # becomes false and we don't use automatically-estimated exposures for the rest of this object
            if self.estimating_exposure:
                # bracket with these values so we can distinguish between user and programmatic exposure setting
                if exposure:
                    self.auto_exposure = True
                    self.exposure = round(exposure)
                    self.auto_exposure = False

            # create sub
            s = Sub(name=name, scope=self.scope, sub_type=self.sub_type, rejected=rejected)

            # add to stack 
            self.scope.stacker.add_sub(s, align_direction=align_direction)

        except Exception as e:
            # we ignore any problems, as sub will still be in watched folder
            Logger.exception('Obj: add_sub')

    def set_name(self, name):
        '''User has changed name field so look it up'''
        self.Name = name.strip()
        if self.Name:
            object_details = self.scope.observinglist.lookup_name(self.Name)
            if object_details:
                Logger.debug('Obj: found dso name {:}'.format(self.Name))                
                for p in self.obj_props:
                    if p != 'Notes':
                        setattr(self, p, str(object_details.get(p, '')))
            else:
                self.scope.notifier.notify('Name not in DSO database')
                Logger.debug('Obj: unable to find dso name {:}'.format(self.Name))                

#============================
# panel animation/pinning
#============================

    def show(self):
        if self.pinned or (self.state in ['displayed', 'incoming']):
            return
        self.anim = Animation(x=0, duration=.25)
        self.anim.bind(on_start=self._incoming, on_complete=self._displayed)
        self.anim.start(self)
           
    def hide(self):
        if self.pinned or (self.state in ['leaving', 'hidden']):
            return
        self.anim = Animation(x=-self.width, duration=.35)
        self.anim.bind(on_start=self._leaving, on_complete=self._hidden)
        self.anim.start(self)

    def on_pinned(self, *args):
        if self.pinned:
            self.state = 'displayed'
        else:
            self.state = 'hidden'
        self.scope.redraw()

    def _displayed(self, *args):
        self.state = 'displayed'
        self.pin.pos = self.width, Window.height - self.pin.height

    def _incoming(self, *args):
        self.state = 'incoming'

    def _leaving(self, *args):
        self.pin.x = -self.pin.width
        self.state = 'leaving'

    def _hidden(self, *args):
        self.state = 'hidden'

