''' Observations. Manages previous observations and associated observations table.
'''

import glob
import os
import json
import random
import shutil
from datetime import datetime

from kivy.logger import Logger
from kivy.event import EventDispatcher
from kivy.core.window import Window
from kivy.uix.boxlayout import BoxLayout

from jocular.utils.component import Component
from jocular.utils.utils import add_if_not_exists, get_fits, RA_hours_to_degrees
from jocular.widgets.table import Table
from jocular.utils.RA_and_Dec import RA, Dec

class Observations(BoxLayout, EventDispatcher, Component):

    def __init__(self, scope, **kwargs):
        self.scope = scope
        super().__init__(**kwargs)

    def load_observations(self):

        Logger.info('Observations: loading previous capture metadata')

        # load all previously-observed objects
        self.observations = {}
        try:
            sessions = glob.glob(os.path.join(self.scope.captures_dir, '*'))
            paths = [glob.glob(os.path.join(s, '*')) for s in sessions]
            for path in [l for subl in paths for l in subl if os.path.isdir(l)]:
                json_path = os.path.join(path, 'info.json')
                name = os.path.basename(path)

                # generate metadata where no info.json
                if not os.path.exists(json_path):
                    Logger.info('Observations: generating metadata for {:}'.format(name))
                    all_fits = get_fits(path)
                    if all_fits:
                        sub_type = 'light'  # the default
                        if len(name) >= 4:
                            if name[:4] in ['dark', 'flat', 'bias']:
                                sub_type = name[:4]
                        hdr = {
                            'Name': name,
                            'sub_type': sub_type,
                            'n_subs': len(all_fits),
                            'session': datetime.fromtimestamp(os.path.getmtime(all_fits[0])).strftime('%d %b %y %H:%M')
                        }
                        with open(json_path, 'w') as f:
                            json.dump(hdr, f, indent=4)

                # now load info.json
                if os.path.exists(json_path):
                    with open(json_path, 'r') as f:
                        self.observations[path] = json.load(f)
                    # in case we are using an alpha version and don't have n_subs, update them
                    if 'n_subs' not in self.observations[path]:
                        Logger.debug('Observations: updating {:}'.format(json_path))
                        self.observations[path]['n_subs'] = len(get_fits(path))
                        if 'selected' in self.observations[path]:
                            del self.observations[path]['selected']
                        with open(json_path, 'w') as f:
                            json.dump(self.observations[path], f, indent=4)

        except Exception as e:
            Logger.exception('Observations: load_observations')

        # convert RA and Dec for obs that have them
        for o, v in self.observations.items():
            if ('RA' in v) & ('Dec' in v):
                self.observations[o]['RA'] = RA(v['RA'])  
                self.observations[o]['Dec'] = Dec(v['Dec'])

    def build_observations(self):

        if not hasattr(self, 'observations'):
            self.load_observations()

        return Table(
            size=Window.size,
            data=self.observations,
            name='Previous observations',
            cols={
                'Name': {'w': 250, 'align': 'left', 'sort': {'catalog':''}, 'action': self.load_dso},
                'OT': {'w': 40},
                'Con': {'w': 50},
                'RA': {'w': 110, 'align': 'right', 'type': float, 'val_fn': RA_hours_to_degrees},
                'Dec': {'w': 110, 'align': 'right', 'type': float},
                'Mag': {'w': 60, 'align': 'right', 'type': float},
                'Diam': {'w': 55, 'align': 'right', 'type': float},
                'Exposure': {'w': 80, 'field': 'exposure', 'type': float},
                'Temp': {'w': 60, 'field': 'temperature', 'type': float},
                'Session': {'w': 160, 'field': 'session', 'sort': {'DateFormat': '%d %b %y %H:%M'}},
                'N': {'w': 40, 'field': 'n_subs', 'type': int},
                'Notes': {'w': 1, 'align': 'left'}
                },
            actions={'move to delete folder': self.move_to_delete_folder},
            on_hide_method=self.scope.table_hiding
            )        

    def show_observations(self, *args):
        '''Called from menu to browse DSOs; open on first use'''

        if not hasattr(self, 'observations_table'):
            self.observations_table = self.build_observations()
        self.scope.showing = 'observations'

        # redraw on demand as it is expensive
        if self.observations_table not in self.scope.children:
            self.scope.add_widget(self.observations_table, index=0)

        self.observations_table.show()    

    def move_to_delete_folder(self, *args):
        for s in self.observations_table.selected:
            if s in self.observations:
                self._delete(s)
        self.observations_table.update()

    def _delete(self, path):
        to_path = self.scope.deleted_captures_dir
        Logger.debug('Observations: moving {:} to {:}'.format(path, to_path))
        add_if_not_exists(to_path)
        dname = os.path.join(to_path,
            '{:}_{:}_{:d}'.format(os.path.basename(path),
            datetime.now().strftime('%d_%b_%y_%H_%M'),
            random.randint(1,9999)))
        try:
            shutil.move(path, dname)
        except Exception as e:
            Logger.exception('Observations: problem moving {:} to {:}'.format(path, dname))

        # remove from observations dict
        if path in self.observations:
            del self.observations[path]

    def rename(self, old_path=None, new_path=None):
        Logger.debug('Observations: renaming {:} to {:}'.format(old_path, new_path))
        if not hasattr(self, 'observations'):
            self.load_observations()
        if old_path and new_path:
            if old_path in self.observations:
                self.observations[new_path] = self.observations[old_path]
                del self.observations[old_path]
                # NEW LINE that might fix odd bug
                self.observations_table.update()

    def load_dso(self, row):
        self.observations_table.hide()
        self.scope.obj.load(row.key)

    def update(self, path, hdr):
        Logger.debug('Observations: adding/updating metadata for {:}'.format(path))
        if not hasattr(self, 'observations'):
            self.load_observations()
        # add nsubs field
        hdr['n_subs'] = 0
        for t in ['selected', 'rejected', 'nonaligned']:
            if t in hdr:
                hdr['n_subs'] += len(hdr[t])
        self.observations[path] = hdr

   