'''Handles DSO database and observing list management
'''

import os
import json

from kivy.logger import Logger
from kivy.core.window import Window
from kivy.properties import BooleanProperty
from kivy.event import EventDispatcher

from jocular.utils.component import Component
from jocular.utils.utils import datenow, convert_RA_Dec, RA_hours_to_degrees
from jocular.utils.RA_and_Dec import RA, Dec
from jocular.widgets.table import Table

class Observinglist(EventDispatcher, Component):

    remove_on_observation = BooleanProperty(False)

    def __init__(self, scope):
        self.settables = self.properties()
        super().__init__()
        self.load_settings(settings=scope.settings)
        self.scope = scope
        self.allowload = False  # determine whether we ignore keypresses (eg in browse versus load mode)

    def load_observing_list(self):
        Logger.debug('ObservingList: loading observing list')
        try:
            with open(os.path.join(self.scope.user_data_path, 'observing_list.json'), 'r') as f:
                self.observing_list = convert_RA_Dec(json.load(f))
        except Exception as e:
            Logger.info('ObservingList: creating empty observing list')
            self.observing_list = {}

    def load_dsos(self):
        # DSO database
        Logger.debug('ObservingList: loading DSO database')
        try:
            with open(os.path.join(self.scope.user_data_path, 'dsos.json'), 'r') as f:
                self.dsos = convert_RA_Dec(json.load(f))
        except Exception as e:
            Logger.info('ObservingList: no file dsos.json')
            self.dsos = {}

    def build_dsos(self):

        if not hasattr(self, 'dsos'):
            self.load_dsos()

        cols={
            'Name': {'w': 250, 'align': 'left', 'sort': {'catalog':''}},
            'OT': {'w': 40},
            'Con': {'w': 50},
            'RA': {'w': 110, 'align': 'right', 'type': float, 'val_fn': RA_hours_to_degrees},
            'Dec': {'w': 110, 'align': 'right', 'type': float},
            'Mag': {'w': 55, 'align': 'right', 'type': float},
            'Diam': {'w': 65, 'align': 'right', 'type': float},
            'SB': {'w': 60, 'align': 'right', 'type': float},
            'Dist': {'w': 90, 'align': 'right', 'sort': {'catalog':''}},
            'Type': {'w': 65, 'align': 'right'},
            'B-V': {'w': 55, 'align': 'right', 'type': float, 'field': 'BV'},
            'N': {'w': 55, 'align': 'right', 'type': int},
            'Notes': {'w': 1, 'align': 'left', 'field': 'Other'}
            }

        return Table(
            size=Window.size,
            data=self.dsos,
            name='DSOs',
            cols=cols,
            update_on_show=False,
            actions={'add to observing list': self.add_to_observing_list},
            on_hide_method=self.scope.table_hiding
            )

    def build_observing_list(self):

        if not hasattr(self, 'observing_list'):
            self.load_observing_list()

        cols={
            'Name': {'w': 250, 'align': 'left', 'sort': {'catalog':''}, 'action': self.new_from_list},
            'OT': {'w': 40},
            'Con': {'w': 50},
            'RA': {'w': 110, 'align': 'right', 'type': float, 'val_fn': RA_hours_to_degrees},
            'Dec': {'w': 110, 'align': 'right', 'type': float},
            'Mag': {'w': 55, 'align': 'right', 'type': float},
            'Diam': {'w': 75, 'align': 'right', 'type': float},
            'Added': {'w': 140, 'sort': {'DateFormat': '%d %b %y %H:%M'}},
            'Notes': {'w': 1, 'align': 'left', 'field': 'Other'}
            }

        return Table(
            size=Window.size,
            data=self.observing_list,
            name='Observing list',
            cols=cols,
            actions={'remove from observing list': self.remove_from_observing_list},
            on_hide_method=self.scope.table_hiding
            )

    def show_observing_list(self, *args):
        '''Called from menu to browse observing list'''
        self.allow_load = False
        self._show_observing_list()

    def new_from_observing_list(self, *args):
        '''Called from menu to load name from observing list'''
        self.allow_load = True
        self._show_observing_list()

    def _show_observing_list(self, *args):
        '''Called from menu to browse observing list; open on first use'''
        if not hasattr(self, 'observing_list_table'):
            self.observing_list_table = self.build_observing_list()
        self.scope.showing = 'observing list'

        # redraw on demand when required
        if self.observing_list_table not in self.scope.children:
            self.scope.add_widget(self.observing_list_table, index=0)

        self.observing_list_table.show()    

    def show_DSOs(self, *args):
        '''Called from menu to browse DSOs; open on first use'''
        if not hasattr(self, 'dsos_table'):
            self.dsos_table = self.build_dsos()
        self.scope.showing = 'DSOs'

        # redraw on demand when required
        if self.dsos_table not in self.scope.children:
            self.scope.add_widget(self.dsos_table, index=0)

        self.dsos_table.show()    

    def on_close(self):
        if hasattr(self, 'observing_list'):
            self.save_observing_list()

    def new_from_list(self, row, *args):
        '''Called when user selects a row in the observing list table. We check if table
        was originally opened with load allowed.
        '''
        if self.allow_load:
            dso = row.fields['Name'].text
            Logger.debug('ObservingList: user selected {:} from observing list'.format(dso))
            self.observing_list_table.hide()
            self.scope.obj.new(dso=dso)
            self.allow_load = False
            if self.remove_on_observation:
                if dso in self.observing_list:
                    Logger.debug('ObservingList: removing {:} from observing list'.format(dso))
                    del self.observing_list[dso]

    def add_to_observing_list(self, *args):
        Logger.debug('ObservingList: adding objects to observing list')
        if not hasattr(self, 'observing_list'):
            self.load_observing_list()
        dn = datenow()
        for s in self.dsos_table.selected:
            self.observing_list[s] = self.dsos[s].copy()
            self.observing_list[s]['Added'] = dn
        self.dsos_table.deselect_all()
        self.save_observing_list()

    def remove_from_observing_list(self, *args):
        Logger.debug('ObservingList: removing objects to observing list')
        for s in self.observing_list_table.selected:
            if s in self.observing_list:
                del self.observing_list[s]
        self.observing_list_table.update()  # since we are viewing
        self.save_observing_list()

    def save_observing_list(self, *args):
        Logger.debug('ObservingList: saving observing list')
        try:
            with open(os.path.join(self.scope.user_data_path, 'observing_list.json'), 'w') as f:
                json.dump(self.observing_list, f, indent=1)
        except OSError:
            Logger.exception('ObservingList: problem saving observing list')
            self.scope.notifier.notify('Problem saving observing list')

    def lookup_name(self, nm, *args):
        Logger.debug('ObservingList: look up name in DSO database')
        if not hasattr(self, 'dsos'):
            self.load_dsos()
        # look up object
        if nm in self.dsos:
            od = self.dsos[nm]
        else: # search thru names
            d = [dd for dd in self.dsos.values() if dd['Name'].lower() == nm.lower()]
            if len(d) == 0:
                return None
            else:
                od = d[0]
        od['RA'] = RA(od['RA'])
        od['Dec'] = Dec(od['Dec'])
        return od
