# The main application 

import sys
import json
import os

from kivy.logger import Logger
from kivy.app import App
from kivy.core.window import Window
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.progressbar import ProgressBar
from kivy.properties import NumericProperty, OptionProperty, BooleanProperty
from kivy.clock import Clock
from kivy.metrics import dp

from jocular.utils.utils import joc_icon
from jocular.widgets.widgets import Menu
from jocular.utils.component import Component
from jocular.components.eyepiece import Eyepiece
from jocular.components.obj import Obj
from jocular.components.observations import Observations
from jocular.components.aligner import Aligner  # .3s
from jocular.components.calibrator import Calibrator
from jocular.components.watcher import Watcher
from jocular.components.stacker import Stacker
from jocular.components.processing import Processing
from jocular.components.configurer import Configurer
from jocular.components.camera import Camera
from jocular.components.notifier import Notifier
from jocular.components.observinglist import Observinglist
from jocular.components.snapshotter import Snapshotter
from jocular.components.firsttime import FirstTime

# Main screen, handling everything that isn't handled by the panels
class Scope(FloatLayout, Component):

    live = BooleanProperty(False)
    showing = OptionProperty('main', options=['main', 'DSOs', 'observing list', 'calibration', 'observations'])
    dimmer = NumericProperty(90)
    redrawing = BooleanProperty(False)
    closing = BooleanProperty(False)
    mouse_over_scope = BooleanProperty(False)
    max_megapixels = NumericProperty(2.5)  # limit size of images it makes sense to process at the moment

    settables = ['dimmer', 'watched_dir', 'max_megapixels']

    def __init__(self, app, base_path, jocular_version, **kwargs):

        Logger.info('Jocular: initialising [v{:}]'.format(jocular_version))

        self.initialising_components = True
        self.app = app

        super().__init__(**kwargs)

        self.base_path = base_path

        # Look for user's data path
        try:
            with open(os.path.join('jocular', 'user_path.json')) as f:
                jocular_settings = json.load(f)
            user_data_path = jocular_settings['user_data_path']
            if not os.path.exists(user_data_path):
                raise FileNotFoundError
            else:
                self.normal_init(user_data_path)
        # no json file, or changed data path, so treat as first time user
        except FileNotFoundError:
            self.add_widget(FirstTime(scope=self))
        # any other error, exit
        except Exception:
            raise

    def normal_init(self, user_data_path, first_time=False, screen=None, *args):
        # continue initialisation after getting hold of user data folder
        # this could be called by __init__ or from FirstTime instance

        # if first time, remove first time screen
        if first_time and (screen is not None):
            self.remove_widget(screen)

        self.user_data_path = user_data_path

        # create necessary folders; these might be overriden by settings
        for d in ['captures', 'calibration', 'simulator', 'watched', 'snapshots', 'deleted_captures', 'deleted_masters']:
            path = os.path.join(self.user_data_path, d)
            setattr(self, d + '_dir', path)
            if not os.path.exists(path):
                try:
                    os.mkdir(path)
                except OSError as e:
                    # any issues creating files at this stage, best to exit
                    Logger.exception('Jocular: cannot add {:}'.format(path))
                    sys.exit()

        # load settings, or create if it has somehow been deleted
        self.settings_path = os.path.join(self.user_data_path, 'settings.json')
        try:
            with open(self.settings_path) as f:
                self.settings = json.load(f)
            self.load_settings(settings=self.settings)
        except:
            self.settings = None
            Logger.info('Jocular: no settings.json file')

        self.components = {
            'notifier': Notifier, 
            'obj': Obj,
            'observations': Observations,
            'snapshotter': Snapshotter,
            'stacker': Stacker,
            'aligner': Aligner,
            'calibrator': Calibrator,
            'camera': Camera,
            'configurer': Configurer,
            'observinglist': Observinglist,
            'watcher': Watcher,
            'eyepiece': Eyepiece,
            'processing': Processing}

        # order of loading matters
        for comp, cl in self.components.items():
            setattr(self, comp, cl(scope=self))

        self.initialising_components = False  # it is now OK to redraw
 
        if first_time:
            self.redraw()

        # start listening to mouse position events etc
        Window.bind(mouse_pos=self.on_mouse_pos)

        self._keyboard = Window.request_keyboard(self._keyboard_closed, self)
        self._keyboard.bind(on_key_down = self._on_keyboard_down)
        self._keyboard.bind(on_key_up = self._on_keyboard_up)

        # post-initialisation of components (need to do it this way to ensure scope is built)
        Clock.schedule_once(self.post_init, 0)

    def post_init(self, *args):

        # do any post-initialisation of any components needing it
        Logger.info('Jocular: starting post-init phase')
        for comp in self.components.keys():
            getattr(self, comp).post_init()

        self.obj.new()

    def _keyboard_closed(self):
        self._keyboard.unbind(on_key_down = self._on_keyboard_down)
        self._keyboard = None

    def on_close(self, *args):
        '''Get scope settings, add in those of each component, the menu, and save'''

        Logger.info('Jocular: starting close-down')
        self.closing = True  # in case any of our processes need this info

        # in case we are closing down without initialising
        if not hasattr(self, 'eyepiece'):
            return

        settings = self.save_settings()

        for comp in self.components.keys():
            getattr(self, comp).on_close()

        for comp in self.components.keys():
            settings.update(getattr(self, comp).save_settings())

        try:
            with open(self.settings_path, 'w') as f:
                json.dump(settings, f, indent=4)
        except OSError:
            Logger.exception('Jocular: cannot save settings to {:}'.format(self.settings_path))

        # delete any empty session and object folders
        # rmdir raises an exception for non-empty dirs
        try:
            for path, dirs, files in os.walk(self.captures_dir):
                if (not files) and (not dirs):
                    os.rmdir(path)
        except Exception as e:
            Logger.exception('Jocular: problem removing empty captures dirs')

    def redraw(self, *args):
        '''Build interface on start and on resize.'''

        if self.initialising_components or self.redrawing:
            return

        Logger.info('Jocular: redrawing')
        self.redrawing = True

        # clear canvas/widgets
        self.clear_widgets()
        self.canvas.clear()

        # determine locations of components and pinned state depending on layout
        components_draw_order = ['eyepiece', 'obj', 'observations', 'stacker', 
            'aligner', 'calibrator', 'camera', 'processing', 
            'configurer', 'observinglist', 'watcher', 'snapshotter', 'notifier']

       # switch order of priority (z-index not working)
        if self.processing.pinned and (not self.obj.pinned):
            proc_ind = components_draw_order.index('processing')
            obj_ind = components_draw_order.index('obj')
            components_draw_order[obj_ind] = 'processing'
            components_draw_order[proc_ind] = 'obj'

        # menu buttons
        self.menu = Menu(
            pos=(Window.width - dp(10), Window.height -  dp(30)),
            text_width = dp(200),
            menu=[
                ('new observation   ' + joc_icon('scope', fs=20), self.obj.new_from_menu),
                ('new from observing list   ' + joc_icon('obslist', fs=20), self.observinglist.new_from_observing_list),
                ('', ''),
                ('clear stack   ' + joc_icon('stack', fs=20), self.obj.clear_stack),
                ('mark for deletion  ' + joc_icon('delete', fs=20), self.obj.delete),
                ('', ''),
                ('earlier observations   ' + joc_icon('previous', fs=20), self.observations.show_observations),
                ('observing list   ' + joc_icon('obslist', fs=20), self.observinglist.show_observing_list),
                ('DSO database   ' + joc_icon('search', fs=20), self.observinglist.show_DSOs),
                ('calibration masters   ' + joc_icon('thermometer', fs=20), self.calibrator.show_calibration_table),
                ('', ''),
                ('quit  ' + joc_icon('close', fs=20), self.close)
               ]
            )
        
        self.add_widget(self.menu)

        for comp in components_draw_order:
            getattr(self, comp).redraw()

        # make a simple progress bar for general use
        self.progress = ProgressBar(size_hint=(None, None))
        self.add_widget(self.progress, index=0)
        self.progress.height = dp(30)
        self.progress.y = -self.progress.height

        self.redrawing = False

        # redisplay what was showing
        if self.showing == 'calibration':
            self.calibrator.show_calibration_table()
        elif self.showing == 'DSOs':
            self.observinglist.show_DSOs()
        elif self.showing == 'observing list':
            self.observinglist.show_observing_list()
        elif self.showing == 'observations':
            self.observations.show_observations() 

        self.dimmer += .01  # force update

    on_size = redraw

    def close(self, *args):
        App.get_running_app().stop()

    def _on_keyboard_down(self, *args):
        pass

    # the start of a key-action method; do it on key up to prevent problems with key held down
    def _on_keyboard_up(self, keyboard, keycode, *args):

        if not self.mouse_over_scope:
            return

        key = keycode[1]

        if hasattr(self.eyepiece, 'nsubs') & (self.eyepiece.nsubs > 0):
            if key == 'right':
                self.eyepiece.selected_sub = min(self.eyepiece.selected_sub + 1, self.eyepiece.nsubs - 1)
            elif key == 'left':
                self.eyepiece.selected_sub = max(0, self.eyepiece.selected_sub - 1)
            elif key == 'up':
                self.eyepiece.toggle_selected_sub()
            elif key == 'down':
                self.eyepiece.toggle_sub_or_stack()
            elif key == 'backspace':
                self.eyepiece.delete_selected_sub()

        # let object or stacker handle escape key press
        if key == 'escape':
            self.obj.escape_pressed()
            self.stacker.escape_pressed()

    def on_mouse_pos(self, *args):

        def within(x1, y1, x2, y2, delta):
            return ((x1 - x2)**2 + (y1 - y2)**2)**0.5 < delta

        x, y = args[1]
        w, h = Window.size

        # flag when mouse is in scope area to prevent key presses interfering with input to obj area
        self.mouse_over_scope = x > self.obj.width

        # ignore mouse when showing a table
        if self.showing != 'main':
            self.mouse_over_scope = False
            return

        enter_tol = dp(100)
        exit_margin = dp(50)

        if within(x, y, w, h, enter_tol):
            self.menu.show()
            return

        if within(x, y, 0, h, enter_tol):
            self.obj.show()
            return

        if within(x, y, w, 0, enter_tol):
            self.processing.show()
            return

        # leaving is more panel-specific
        if x > self.obj.width + exit_margin:
            self.obj.hide()

        if y > self.processing.height + exit_margin:
            self.processing.hide()

        if not within(x, y, w, h, dp(320)):
            self.menu.hide()

        # handle configuration pane: we flag first entry then close on exit

        if x < self.configurer.width - dp(50):
            self.configurer.enter()

        elif (x > self.configurer.width + dp(50)) and (y > self.processing.height):
            self.configurer.hide()

    def table_hiding(self, *args):
        self.showing = 'main'

    def on_dimmer(self, *args):
        self.app.dimmed_transp = self.dimmer/100.
 
