'''Snapshotter: produces png/jpg image in eyepiece or landscape style with user-defined
degree of annotation. Called on save if autosave on, and when user hits the snap button.
'''

import os
from datetime import datetime
import numpy as np
from skimage.io import imsave
from PIL import Image, ImageDraw, ImageFont
from astropy.io import fits 

from kivy.logger import Logger
from kivy.core.window import Window
from kivy.properties import OptionProperty, BooleanProperty
from kivy.event import EventDispatcher

from jocular.utils.component import Component
from jocular.utils.utils import purify_name

otypes = {'GC': 'globular', 'GX': 'galaxy', 'PN': 'planetary nebula',
    'G+': 'galaxy group', 'GG': 'galaxy cluster', 'DN': 'dark nebula',
    'BN': 'bright nebula', 'RN': 'reflection nebula', 'QS': 'quasar',
    'BS': 'star', 'OC': 'open cluster',
    'G2': 'G2V star', 'CS': 'carbon star', 'LT': 'L-type star', 'WT': 'W-type star',
    'TT': 'T-type star', 'YT': 'Y-type star', 
    'BQ': 'bright quasar', 'DQ': 'high redshift quasar',
    'SN': 'supernova remnant'}

import mss

class Snapshotter(EventDispatcher, Component):

    snapshot_annotation = BooleanProperty('full', 
        options=['none', 'name only', '+object details', '+capture', '+kit', 'full'])
    save_format = OptionProperty('jpg', options=['png', 'jpg', 'fits'])
    save_style = OptionProperty('eyepiece', options=['eyepiece', 'landscape'])

    def __init__(self, scope):

        self.settables = self.properties()
        self.scope = scope
        super().__init__()
        self.load_settings(settings=scope.settings)

    def mss_grabber(self, bbox=None):
        with mss.mss() as sct:
            im = sct.grab(bbox)
            return Image.frombytes('RGB', im.size, im.bgra, 'raw', 'BGRX').convert('RGBA')

    def save(self, *args):

        def large_font(w):
            f = max(12, int(w / 36))
            return ImageFont.truetype(os.path.join(
                self.scope.base_path, 'jocular', 'resources', 'Roboto-Medium.ttf'), f), f

        def small_font(w):
            f = max(10, int(w / 56))
            return ImageFont.truetype(os.path.join(
                self.scope.base_path, 'jocular', 'resources', 'Roboto-Medium.ttf'), f), f

        if self.scope.eyepiece.image is None:
            return

        obj = self.scope.obj

        # handle fits here as we don't need to go thru anything else
        if self.save_format == 'fits':
            try:
                path = os.path.join(self.scope.snapshots_dir, purify_name(obj.Name))
                nm = '{:} {:}.fits'.format(path, datetime.now().strftime('%d%b%y_%H_%M_%S'))
                hdu = fits.PrimaryHDU(self.scope.eyepiece.image)
                hdu.writeto(nm, overwrite=True)
                Logger.debug('Snapshotter: saved fits as {:}'.format(nm))
                self.scope.notifier.notify('Saved as FITS')
            except Exception as e:
                Logger.debug('Snapshotter: problem saving fits {:}'.format(e))
                self.scope.notifier.notify('Problem saving as FITS')
            return

        black = 20, 20, 20, 255
        grey = 120, 120, 120, 255

        Logger.info('Snapshotter: saving {:} image in {:} style'.format(self.save_format, self.save_style))

        if obj.OT and (obj.OT in otypes):
            self.OT_CON = '{:} in {:}'.format(otypes[obj.OT], obj.Con)
        else:
            self.OT_CON = ''

        sqmtemp = []
        if not obj.SQM_inactive:
            sqmtemp.append('{:5.2f} SQM'.format(obj.SQM))
        if not obj.temperature_inactive:
            sqmtemp.append('{:2d}\N{DEGREE SIGN}C'.format(obj.temperature))
        self.SQM_TEMP = ', '.join(sqmtemp)

        kit = {
            'telescope': '{:}', 
            'camera': '{:}', 
            'other_kit': '{:}'}

        obj_details2 = {
            'Dist': '{:}', 
            'Mag': 'mag {:}', 
            'Diam': "{:}'"}

        obj_details1 = {
            'OT_CON': '{:}', 
            'RA': 'RA {:}', 
            'Dec': 'Dec {:}'}

        obj_details_both = {
            'OT_CON': '{:}', 
            'RA': 'RA {:}', 
            'Dec': 'Dec {:}',            
            'Dist': '{:}', 
            'Mag': 'mag {:}', 
            'Diam': "{:}'"
            }

        session_details = {
            'session': '{:}', 
            'SQM_TEMP': '{:}', 
            'seeing': 'seeing {:}', 
            'transparency': 'transparency {:}'}

        processing_details = self._processing_block()

        level = self.snapshot_annotation
        
        if self.save_style == 'eyepiece':

            cx = int(Window.left + self.scope.eyepiece.origin[0])
            cy = int(Window.top + self.scope.height - self.scope.eyepiece.origin[1])
            r = int(self.scope.eyepiece.radius - 40)
            w = 2 * r

            try:
                im = self.mss_grabber(bbox=(cx - r, cy - r, cx + r, cy + r))
            except Exception as e:
                Logger.exception('SnapShotter: problem with image grab')

            blur = 17                     # radius of blurred edge in pixels

            # construct alpha channel to composite eyepiece circle
            # quite likely some of these steps are redundant
            alpha = Image.new('L', (w, w), color=0)
            _alpha = np.ones((w, w))
            grid_x, grid_y = np.meshgrid(np.arange(w), np.arange(w))
            d = pow((grid_x - r) ** 2 + (grid_y - r) ** 2, .5) - (r - 30)
            db = d <= blur
            _alpha[db] = d[db] / blur
            alpha.putdata(_alpha.ravel(), scale=255)
            grad_im = Image.new('RGBA', (w, w), color=(255, 255, 255)) 
            grad_im.putalpha(alpha)

            # composite with scope image, converting if needed (to test)
            try:
                if im.mode != 'RGBA':
                    im = im.convert('RGBA')

                im = Image.alpha_composite(im, grad_im)
            except Exception as e:
                Logger.exception('Snapshotter: problem compositing')
                self.scope.notifier.notify('Snapshotter: problem creating snapshot')
                return

            # annotate Draw object with required level of detail
            sf, sf_size = small_font(w)
            lf, lf_size = large_font(w)
            rowsep = sf_size + 6
            if level != 'none':    
                draw = ImageDraw.Draw(im)
                w, h = im.size
                row = self._draw_aligned(draw, 0, 0, w, obj.Name, lf, black, align='left', rowsep=lf_size+30)
                if level != 'name only':
                    self._show_block(obj_details_both, draw, row, 0, w, sf, grey, align='left', rowsep=rowsep)
                    if level != '+object details':
                        self._show_block(processing_details, draw, h - rowsep, 0, w, sf, grey, align='right', rowsep=-rowsep)
                        if level != '+capture':
                            self._show_block(kit, draw, h - rowsep, 0, w, sf, grey, align='left', rowsep=-rowsep) # neg builds from base
                            if level != '+kit':
                                self._show_block(session_details, draw, 5, 0, w, sf, grey, align='right', rowsep=rowsep)

        elif self.save_style == 'landscape':

            # to do: better scaling of fonts based on image size
            nm = '_tmp.png'

            try:
                imsave(nm, self.scope.eyepiece.scatter.last_image)
                orig_im = Image.open(nm)
            except Exception as e:
                Logger.exception('Snapshotter: problem with image save/load')

            w, h = orig_im.size
            if w < 500:
                wfac = .8
            elif w > 1000:
                wfac = 1.5
            else:
                wfac = 1.
            if level == 'none':
                new_height = h
            elif level == 'name only':
                new_height = int(h + 40 * wfac)
            else:
                new_height = int(h + 115 * wfac)
            im = Image.new('RGBA', (w, new_height), color=(255, 255, 255)) 
            
            im.paste(orig_im, (0, 0, w, h))
            draw = ImageDraw.Draw(im)

            # NB this is indep of screen size, unlike eyepiece view 
            sf, sf_size = small_font(w)
            lf, lf_size = large_font(w)
            rowsep = sf_size + 6

            if level == 'none':
                pass
            elif level == 'name only':
                row = self._draw_aligned(draw, h, 0, w, obj.Name, lf, black, align='center', rowsep=33 * wfac)
            else:
                row = self._draw_aligned(draw, h, 0, w, obj.Name, lf, black, align='center', rowsep=33 * wfac)
                draw.line((5, row, w - 5, row), fill=grey, width=1)
                row += 3
                cols = [int(x) for x in w*np.linspace(0, 1, 6)]
                self._show_block(obj_details1, draw, row, cols[2], cols[3], sf, grey, align='center', rowsep=rowsep)
                if level != '+object details':
                    self._show_block(processing_details, draw, row, cols[4], cols[5], sf, grey, align='right', rowsep=rowsep) # neg builds from base
                    if level != '+capture':
                        self._show_block(kit, draw, row, cols[0], cols[1], sf, grey, align='left', rowsep=rowsep) # neg builds from base
                        if level != '+kit':
                            self._show_block(session_details, draw, row, cols[3], cols[4], sf, grey, align='center', rowsep=rowsep)
                            self._show_block(obj_details2, draw, row, cols[1], cols[2], sf, grey, align='center', rowsep=rowsep)

            # delete temporary file
            os.remove(nm)

        # save and clean up
        self._save(im, obj.Name)
        Logger.info('Snapshotter: saved')
        self.scope.notifier.notify('saved snapshot')


    def _draw_aligned(self, draw, row, l, r, text, font, fill, margin=5, rowsep=20, align='center'):
        # helper to draw pixel text aligned; return new row position
        # alignment is within range of cols, a pair (l, r)
        textwidth, _ = draw.textsize(text, font)
        if align == 'left':
            col = l + margin
        elif align == 'right':
            col = r - textwidth - margin
        elif align == 'center':
            col = (l + r - textwidth) / 2
        draw.text((col, row), text, font=font, fill=fill)
        return row + rowsep

    def _show_block(self, block, draw, row, l, r, font, fill, align='center', rowsep=20):

        for param, fmt in block.items():
            if hasattr(self.scope.obj, param):
                v = getattr(self.scope.obj, param)
            elif hasattr(self, param):
                v = getattr(self, param)
            else:
                v = None
            if (type(v) == str) and (v == 'not set'):
                v = None
            if v:
                row = self._draw_aligned(draw, row, l, r, fmt.format(v), font, fill, align=align, rowsep=rowsep)

    def _processing_block(self):

        block = {}
        if self.scope.eyepiece.sub_or_stack == 'sub':
            nsubs = 1
        else:
            nsubs = len([s for s in self.scope.stacker.subs[:self.scope.eyepiece.selected_sub+1] if s.status == 'select'])
        sub_stats = '{:d}x{:2.0f}s'.format(nsubs, self.scope.obj.exposure)
        mins, secs = divmod(round(nsubs * self.scope.obj.exposure), 60)
        if mins == 0:
            total_exposure = '{:}s'.format(secs)
        elif secs == 0:
            total_exposure = '{:}m'.format(mins)
        else:
            total_exposure = '{:d}m {:d}s'.format(mins, secs)
        self._exposure_details = '{:} ({:})'.format(total_exposure, sub_stats)
        block['_exposure_details'] = '{:}'

        self._stack = self.scope.eyepiece.stack_combine_method
        if nsubs == 1:
            block['_stack'] = 'single sub'
        else:
            block['_stack'] = '{:} stack'
        self._stretch = self.scope.eyepiece.stretch
        block['_stretch'] = '{:} stretch'

        self._calib = ', '.join(self.scope.calibrator.masters.keys())
        block['_calib'] = '{:}'

        return block

    def _save(self, im, obj_name):
        # save, handling any format conversions
        try:
            path = os.path.join(self.scope.snapshots_dir, purify_name(obj_name))
            nm = '{:} {:}.{:}'.format(path, datetime.now().strftime('%d%b%y_%H_%M_%S'), self.save_format)
            if self.save_format == 'png':
                im.save(nm)
            elif self.save_format == 'jpg':
                bg = Image.new('RGB', im.size, (255, 255, 255))
                bg.paste(im, im)
                bg.save(nm)
        except Exception as e:
            Logger.exception('Snapshotter: _save')



