# Stacker. Manages calibration, alignment and stacking of incoming subs

import numpy as np

from kivy.properties import BooleanProperty, NumericProperty
from kivy.event import EventDispatcher
from kivy.clock import Clock
from scipy.stats import trimboth
from jocular.utils.component import Component
from kivy.logger import Logger

class Stacker(EventDispatcher, Component):

    update_time = NumericProperty(0)
    force_update = BooleanProperty(False)   # probably doesn't need to be a property

    def __init__(self, scope, **kwargs):
        self.scope = scope
        super().__init__(**kwargs)
        self.load_settings(settings=scope.settings)
        self.reset()

    def post_init(self, *args):
        self.params = {
            #'eyepiece': ['binning'],
            'calibrator': ['apply_BPM', 'bpm_frames', 'apply_dark', 'apply_flat',
                'current_dark', 'current_flat', 'current_bias'],
            'aligner': ['star_intensity', 'min_sigma', 'max_sigma', 'centroid_method', 
                'do_align', 'min_stars', 'ransac_min_samples', 'ransac_max_trials', 
                'ransac_residual_threshold', 'min_inliers']
            }
        self.current_params = self.get_params()

    def reset(self):
        self.subs = []
        self.selected = []
        self.last_method = ''
        self.exposure = 0
        self.calibration_time = 0
        self.align_time = 0
        self.dont_check_recompute = False

    def is_empty(self):
        return len(self.subs) == 0

    def shape(self):
        if self.is_empty():
            return None
        if self.subs[0].image is None:
            return None
        w, h = self.subs[0].image.shape
        return [w, h]

    # def binning(self):
    #     # helper method for eyepiece
    #     return self.current_params['binning']

    def add_sub(self, sub, align_direction='forward'):
        # add new sub to stack

        try:
            # if this is first sub on stack, ensure we update calibration
            # Done in sub.process as it relies on image shape and hence image being loaded
            first_sub = len(self.subs) == 1
            if first_sub:
                self.current_params = self.get_params()
            sub.process(params=self.current_params, align_direction=align_direction, 
                update_calibration=first_sub, first_sub=first_sub)
            self.subs.append(sub)
            self.scope.eyepiece.sub_added()
        except FileNotFoundError:
            # ignore it, but don't allow it to join stack
            pass

    def delete_selected_sub(self, selected_sub):
        # delete current sub if possible

        sub = self.subs[selected_sub]
        if sub.keyframe:
            Logger.debug('Stacker: Cannot delete key frame')
            self.scope.notifier.notify('Cannot delete key frame (yet!)')
            raise

        try:
            sub.delete()
            del self.subs[selected_sub]
            # to do: update warp model
        except Exception as e:
            Logger.debug('Stacker: delete_selected_sub problem {:}'.format(e))
            self.scope.notifier.notify('Unable to delete sub')
            raise

    def get_params(self):
        # any change to any of these params has potential to require a selective update
        current_params = {}
        for component, ps in self.params.items():
            for p in ps:
                current_params[p] = getattr(getattr(self.scope, component), p)
        return current_params

    def check_recompute(self, *args):
        '''Determine if an update is required. Called on any event that could potentially
        lead to an out-of-date sub or process.
        '''

        # there are situations where we want to override the check_recompute e.g.
        # when doing adaptive star alignment
        if self.dont_check_recompute or (len(self.subs) == 0):
            return

        # don't check for updates for calibration subs
        if self.subs[0].sub_type != 'light':
            return

        # check if any change
        latest_params = self.get_params()
        if self.current_params == latest_params:
            self.update_time = 0
        else:
            self.update_time = sum([s.timing for s in self.subs])

    def realign_from_current(self):
        pass
        #self.scope.obj.keyframe = self.subs[self.scope.selected_sub].name
        #self.recompute()

    def recompute(self):
        '''Called when user requests recompute.
        '''

        # set up progress bar
        pb = self.scope.progress
        pb.y = 0
        pb.size = self.scope.processing.width, 20
        pb.max = len(self.subs)
        pb.value = 0
        self.count = 0

        self.scope.aligner.reset() # resets keyframe etc

        # consider saving exposure time since we are stopping watcher?
        if self.scope.live:
            self.scope.watcher.stop()

        self.dont_check_recompute = True
        self.stop_loading = False

        # load each one with a clock call
        Clock.schedule_once(self.load_from_queue, 0.01)

    def escape_pressed(self):
        self.stop_loading = True

    def load_from_queue(self, dt):

        if self.count < len(self.subs):
            first_sub = self.count == 0
            if first_sub:
                self.current_params = self.get_params()
            self.subs[self.count].process(params=self.current_params, first_sub=first_sub)
            self.subs[self.count].procs = []  # done processes
            self.count += 1
            self.scope.progress.value += 1
            if not self.stop_loading:
                Clock.schedule_once(self.load_from_queue, 0)
        else:
            self.force_update = True  # ensure that we update stack
            self.scope.eyepiece.on_selected_sub()
            self.force_update = False
            self.update_time = 0
            self.scope.progress.y = -self.scope.progress.height
            self.dont_check_recompute = False

            # restart watcher if in live mode
            if self.scope.live:
                self.scope.watcher.start()

    def reorder_subs_sequence(self, names, keyframe=None):
        '''Order subs so they start with the keyframe, then work back
        to start, then from keyframe to end, in order to support
        alignment from an arbitrary keyframe. Returns subs name sequence
        and list of alignment directions.
        '''
        
        if keyframe and (keyframe in names):
            keyloc = names.index(keyframe)
        else:
            keyloc = 0

        if keyloc == 0:
            # all forward
            align_dirs = ['forward' for i, _ in enumerate(names)]
        elif keyloc == len(names) + 1:
            # last sub
            names = names[::-1]
            align_dirs = ['backward' for i, _ in enumerate(names)]            
        else:
            # keyframe is in the middle somewhere
           names = names[keyloc::-1] + names[keyloc+1:]
           align_dirs = ['backward' if i < keyloc else 'forward' for i, _ in enumerate(names)]

        return names, align_dirs

    def compute_stack(self, up_to=None):
        # compute stack on demand
        # Check if anything has changed, recomputing stack if necessary
        # Returning None implies no change in stack or no stack; caller can decide
        # on action in this case (usually do nothing)

        def _percentile_clip(a, perc=80):
            return np.mean(trimboth(np.sort(a, axis=0), (100 - perc)/100, axis=0), axis=0)
 
        # find indices of subs to use (selected, and up to given limit, or all)
        if up_to is None:
            selected = [i for i, s in enumerate(self.subs) if s.status=='select']
        else:
            selected = [i for i, s in enumerate(self.subs[:up_to+1]) if s.status=='select']

        # if none selected, return None
        if not selected:
            return None

        # if same selection and same combine method, return None (nothing has changed)
        if set(selected) == set(self.selected) and \
            (self.scope.eyepiece.stack_combine_method == self.last_method):
                if not self.force_update:
                    return None

        self.last_method = 'mean'
        self.selected = selected

        # 1 selected, so return it
        if len(selected) == 1:
            return self.subs[selected[0]].image

        # create stack
        stk = np.stack([s.image for i, s in enumerate(self.subs) if i in selected])

        # if fewer than 3, use mean
        if len(selected) < 3:
            return np.mean(stk, axis=0)

        method = self.scope.eyepiece.stack_combine_method
        self.last_method = method

        if method == 'mean':
            return np.mean(stk, axis=0)
        elif method == 'median':
            return np.median(stk, axis=0)
        elif method == '70%':
            return _percentile_clip(stk, perc=70)
        elif method == '80%':
            return _percentile_clip(stk, perc=80)
        elif method == '90%':
            return _percentile_clip(stk, perc=90)

    # progressive mean (not in inspect mode) conceptually is simply:
    # stack = (N*stack at frame N + k * stack at frame N+1) / N + k
    # where k is the fraction of time elapsed between frame N and N + 1
    # In practice, implemented by updating the stack in each sub-frame interval
    # set framecount = N-1 and choose interval between updates [0-1] M
    # then framecount += M
    #      stack = (stack + M*sub_N) / framecount
    # so eg if M = 1 we get normal behaviour for mean


