''' Monitor. Looks for FITs arriving in the watched folder, identifies type, and 
passing them on to Obj for further handling.
'''

import os
import numpy as np
import time

from kivy.logger import Logger
from kivy.event import EventDispatcher
from kivy.clock import Clock

from jocular.utils.component import Component
from jocular.utils.utils import get_fits

class Watcher(EventDispatcher, Component):

    def __init__(self, scope):

        super(Watcher, self).__init__()
        self.watching_event = None
        self.scope = scope

    def post_init(self, *args):
        if self.scope.live:
            self.start()

    def start(self):
        Logger.info('Watcher: started')
        self.already_seen = set({})
        if self.watching_event is None:
            # don't monitor too often as not currently supporting shorts
            self.watching_event = Clock.schedule_interval(self.watch, .25)
        self.read_times = []

    def stop(self):
        Logger.info('Watcher: stopped')
        if self.watching_event is not None:
            self.watching_event.cancel()
            self.watching_event = None

    def on_close(self):
        self.stop()
 
    def watch(self, dt):
        '''Monitor watched folder. To avoid reading before generating process has finished, catch 
        exceptions and revisit on next Clock cycle. There might be some exceptions thrown as a result 
        eg array too small, but we ignore them.
        '''

        for nm in [os.path.basename(f) for f in get_fits(self.scope.watched_dir)]:
            if nm in self.already_seen:
                Logger.debug('Watcher: {:} already processed'.format(nm))
            else:
                try:
                    Logger.debug('Watcher: new sub ' + nm)
                    self.already_seen.add(nm)
                    self.read_times.append(time.time())

                    # handles SLL file naming format
                    sub_type = 'light'
                    if nm.startswith('Image_') and len(nm) > 10:
                        if nm[6:10].lower() in ['flat', 'dark', 'bias']:
                            sub_type = nm[6:10].lower()

                    elif len(nm) >= 4:
                       if nm[:4].lower() in ['flat', 'dark', 'bias']:
                            sub_type = nm[:4].lower()

                    self.scope.obj.add_sub(
                        name=nm, 
                        from_dir=self.scope.watched_dir, 
                        sub_type=sub_type, 
                        exposure=self.estimated_exposure())
               
                except Exception as e:
                    Logger.exception('Watcher: watch')


    def reset_exposure_estimate(self):
        '''Called by Obj on getting a first sub in live mode. If we already
        had a single sub we preserve last read time, otherwise set to empty.
        '''
        if len(self.read_times) == 1:
            self.read_times = [self.read_times[0]]
        else:
            self.read_times = []


    def estimated_exposure(self):
        '''Estimate exposure based on sub arrival times
        '''

        if len(self.read_times) < 2:
            return 0

        deltas = np.diff(np.sort(self.read_times))

        if len(deltas) < 3:
            est = np.mean(deltas)
        else:
            est = np.median(deltas)

        # disallow crazy estimates
        if est > 100:
            est = 0

        est = max(1, int(np.round(est))) 

        Logger.debug('Watcher: estimated exposure {:2d}'.format(est))

        return est          
       
