''' sub.py represents a single sub of any given type
Marshalls all processing on a sub (loading FITs, BPM, calibration, etc)
'''

import os
import time
import shutil

from skimage.transform import warp
from astropy.io import fits 

from kivy.logger import Logger
from jocular.utils.utils import add_if_not_exists

class Sub:

    def __init__(self, name=None, scope=None, sub_type=None, rejected=False):

        self.image = None
        self.sub_type = sub_type
        self.name = name
        self.status = 'reject' if rejected else 'select'
        self.message = ''
        self.keyframe = False
        self.scope = scope
        self.warp_model = None
        self.centroids = []     
        #self.procs = {}     # process parameters
        self.timing = 0      # process times eg centroiding, alignment

    def start_timer(self):
        self._t = time.time()
        self.t0 = time.time()

    def elapsed(self, reset=True):
        _t = self._t
        if reset:
            self._t = time.time()
        return int(1000 * (time.time() - _t))

    def total_time(self):
        return time.time() - self.t0

    def delete(self):
        # sub moves itself to deleted folder
        delete_dir = os.path.join(self.scope.obj.save_dir, 'deleted')
        try:
            add_if_not_exists(delete_dir)
            shutil.move(os.path.join(self.scope.obj.save_dir, self.name),  delete_dir)
            Logger.debug('Sub: deleted sub {:}'.format(self.name))
        except OSError as e:
            Logger.debug('Sub: problem moving sub to delete folder {:} {:}'.format(delete_dir, e))
            raise

    def load_image(self, nm):
        '''Loads FITS format file. Uses astropy (which is faster for loading the initial sub)
        rather than skimage simply because the latter takes a couple of seconds longer on the first sub. 
        Return None if any kind of problem e.g. this is where we limit the size of any images the user might 
        wish to use.
        '''

        if not os.path.exists(nm):
            return None

        # check if it has fit or fits extension
        ext = os.path.splitext(nm)[-1].lower()
        if ext not in ['.fit', '.fits']:
            return None

        try:
            with fits.open(nm) as hdu1:
                hdr = hdu1[0].header

                if 'BITPIX' in hdr:
                    if hdr['BITPIX'] != 16:
                        Logger.debug('Sub: Not 16 bits {:}'.format(hdr['BITPIX']))
                        self.scope.notifier.notify('Can only handle 16-bit images at present.')
                        return None
                       
                if 'NAXIS' in hdr:
                    if int(hdr['NAXIS']) != 2:
                        Logger.debug('Sub: load_image: NAXIS not 2 {:}'.format(int(hdr['NAXIS'])))
                        return None

                if ('NAXIS1' in hdr) and ('NAXIS2' in hdr):
                    imsize = int(hdr['NAXIS1']) * int(hdr['NAXIS2']) / 10 ** 6
                    if imsize > self.scope.max_megapixels:
                        Logger.info('Sub: FITS image too large. Ignoring.')
                        self.scope.notifier.notify('Image too large {:3.1f}m  [max={:3.1f}m]'.
                            format(imsize, self.scope.max_megapixels))
                        return None

                # read data and then make checks again in case hdr keywords didn't exist
                im = hdu1[0].data

                if im.ndim != 2:
                    Logger.debug('Sub: image not 2-dim {:}'.format(im.ndim))
                    self.scope.notifier.notify('Can only handle 2-D images at present.')
                    return None

                w, h = im.shape
                imsize = w * h / 10 ** 6
                if imsize > self.scope.max_megapixels:
                    Logger.info('Sub: FITS image too large. Ignoring.')
                    self.scope.notifier.notify('Image too large {:3.1f}m  [max={:3.1f} m'.
                        format(imsize, self.scope.max_megapixels))
                    return None

                im = im / 2 ** 16

                return im

        except Exception as e:
            Logger.exception('Sub: problem loading FITS {:}'.format(nm))
            return None

    def process(self, params=None, align_direction='forward', update_calibration=False, first_sub=False):
        '''Full (Re)process of sub to bring it up to date. We use the processing parameters
        defined by the stacker to ensure consistency across the stack.
        Optionally update calibration (usually on the first sub); and if this is the
        first sub, inform star extraction stage in case we are adapting the star threshold. 
        Called by stack when a sub is added.
        '''

        if self.scope.closing:
            return

        self.start_timer()

        # load image
        self.image = self.load_image(os.path.join(self.scope.obj.save_dir, self.name))
        if self.image is None:
            self.scope.notifier.notify('Cannot load FITs')
            Logger.debug('Sub: cannot load {:}'.format(self.name))
            raise FileNotFoundError
            #return
        Logger.debug('Sub: loaded in {:d}'.format(self.elapsed()))

        '''An issue: although we are employing a specified set of params on the 
        stack, if the user changes them during the process, if those changes
        are felt within a component like aligner directly (eg gaussian) then the stack
        will not be consistent. Implication is that setting these params should be
        external to the component itself and they are brought in as params when called.
        This requires a change to config etc too. Needs more thinking about. What
        happens at present is that any changes to these params will be noticed and
        recompute will be highlighted, so the user will see that the stack is not
        up to date or consistent. I guess in keeping with routines like alignment
        being generally useful *outside* Kivy it makes sense to pass in relevant
        parameters. Need to study on a case-by-case basis as it could be different for
        calibration.
        '''

        # calibration sub processing
        if self.sub_type != 'light':
            if self.sub_type == 'flat':
                self.image = self.scope.calibrator.calibrate_flat(self.image)
            if params['apply_BPM']:
                self.image = self.scope.calibrator.process_bpm(self.image)
            return      

        # update masters on the first sub of a stack once we have image dimensions
        # but need to prevent it checking recompute?
        if update_calibration:
            self.scope.calibrator.update_masters()

        # bad pixel mapping
        if params['apply_BPM']:
            self.image = self.scope.calibrator.process_bpm(self.image)
            Logger.debug('Sub: BPM in {:d}'.format(self.elapsed()))

        # apply ROI
        # if hasattr(self.scope.eyepiece, 'ROI'):
        #     if  self.scope.eyepiece.apply_ROI & (self.scope.eyepiece.ROI is not None): 
        #         left, bottom, right, top = self.scope.eyepiece.ROI
        #         self.image = self.image[bottom:top, left:right]
        #         Logger.debug('Sub: ROI in {:d}'.format(self.elapsed()))

        # # apply binning
        # if params['binning'] > 1.2:
        #     self.image = do_binning(self.image, params['binning'])
        #     Logger.debug('Sub: binning in {:d}'.format(self.elapsed()))

        # calibration
        if  params['apply_flat'] or params['apply_dark']:
            self.image = self.scope.calibrator.calibrate_light(self.image)
            Logger.debug('Sub: calibration in {:d}'.format(self.elapsed()))
        
        if not params['do_align']:
            # if previously marked as non-aligned, change to selected since we are not aligning anyway
            # but don't change status of any user-rejected subs
            if self.status == 'nalign':
                self.status = 'select'
            self.timing = self.total_time()
            return

        # extract stars
        # we pass in first_sub flag since we may be adaptively finding star threshold
        raw_stars = self.scope.aligner.extract_stars(self.image, first_sub=first_sub)
        nstars = len(raw_stars)

        # do we have enough stars?
        if len(raw_stars) < params['min_stars']:
            Logger.info('Sub: not enough stars to align {:d}'.format(nstars))
            self.scope.notifier.notify('Insufficient stars to align {:d}'.format(nstars))
            self.timing = self.total_time()
            return

        # compute centroids
        self.centroids = self.scope.aligner.compute_centroids(self.image, raw_stars)
        Logger.debug('Sub: extracted {:d} stars in {:d} ms'.format(nstars, self.elapsed()))

        # align
        self.warp_model, success = self.scope.aligner.align(
            self.centroids, align_direction=align_direction)
        Logger.debug('Sub: align successful? {:} in {:d} ms'.format(success, self.elapsed()))
 
        # warp
        if self.warp_model is None:  # this is the start of the process, hence no warp model
            self.keyframe = True
        elif success:
            #self.keyframe = False
            self.image = warp(self.image, self.warp_model, order=3, preserve_range=True)
            Logger.debug('Sub: warped in {:d} ms'.format(success, self.elapsed()))
            # if previously non-aligned, set to select
            if self.status == 'nalign':
                self.status = 'select'
        else:
            self.status = 'nalign'

        self.timing = self.total_time()

