''' gradient.py simple estimation of 2d linear gradient surface
* consider caching or speeding up for large images
'''

import numpy as np
from numpy.polynomial import polynomial

def estimate_gradient(im, grid_x, grid_y):

    def _polyfit2d(x, y, f, deg):
        deg = np.asarray(deg)
        vander = polynomial.polyvander2d(x, y, deg)
        vander = vander.reshape((-1, vander.shape[-1]))
        f = f.reshape((vander.shape[0],))
        c = np.linalg.lstsq(vander, f, rcond=None)[0]
        return c.reshape(deg+1)

    if im is None:
        return None

    try:
        # fit at random pixels
        npts = 500
        r, c = im.shape

        x = np.random.random_integers(0, high=c-1, size=(npts))
        y = np.random.random_integers(0, high=r-1, size=(npts))
        z = np.array([im[j, i] for i, j in zip(x, y)])

        # remove outliers (likely to be stars)
        zlo, zhi = np.percentile(z, 20), np.percentile(z, 80)
        inds = [i for i, j in enumerate(z) if (j > zlo) and (j < zhi)]

        if len(inds) < 5:
            return None

        # perform fit
        m = _polyfit2d(x[inds], y[inds], z[inds], deg=[2, 1])

        # return gradient evaluated on image mesh
        g = polynomial.polyval2d(grid_x, grid_y, m)
        
        mean_g = np.mean(g)
        return g - mean_g, mean_g, np.std(g)

    except Exception as e:
        return None

