# utils.py

# for next version, reorganise to separate out file-handling code

import os
import glob
import os.path
import math
import operator
from datetime import datetime
#from skimage.transform import resize, rescale

from kivy.logger import Logger
from kivy.vector import Vector

from jocular.utils.RA_and_Dec import RA, Dec

# Jocular font mapping
joc_icons = {
    'pinned': 'p',
    'unpinned': 'u',
    'pause': '3',
    'play': 'n',
    'play stack': 'P',
    'realign': 'Z',
    'oculus': 'o',
    'last_sub': 'f',
    'first_sub': 'e',
    'next_sub': '>',
    'prev_sub': '<',
    'fine': 'l',
    'left arrow': 'z',
    'right arrow': 'A',
    'camera': 'b',
    'reticule': 'V',
    'reset': 'y',
    'scope': 'H',
    'obslist': ':',
    'previous': '{',
    'info': 'o',
    'search': 'a',
    'thermometer': 'I',
    'stack': 'E',
    'delete': '0',
    'close': 'Q',
    'config': 'x',
    'tick on': 'X',
    'tick off': 'Y',
    'ROI': 'd',
    'recompute': 'h',
    'switch down': '1',
    'switch up': '2'
    }

def joc_icon(nm, fs=16):
    fsize = str(fs) + 'sp'
    return "[font=Jocular][size={:}]{:}[/size][/font]".format(fsize, joc_icons.get(nm, 'm'))

# standardised date/time handling

date_time_format = '%d %b %y %H:%M'

def datenow():
    return datetime.now().strftime(date_time_format)

def date_to_int(mydate):
    return datetime.strptime(mydate, date_time_format)

def hours_since_date(mydate):
    return (datetime.now() - date_to_int(mydate)).total_seconds()/3600

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def remove_nulls_from_dict(d):
    return {k: v for k, v in d.items() \
        if not((isinstance(v, str) and v.strip() == '') or (isinstance(v, float) and math.isnan(v)))}

# convert dict to RA/Decetc
def convert_RA_Dec(d):
    for k in d.keys():
        d[k]['RA'] = RA(d[k]['RA'])
        d[k]['Dec'] = Dec(d[k]['Dec'])
    return d

def RA_hours_to_degrees(x):
    return x * 15

# image manipulations
# def do_binning(im, binfac, original_size=False):
#     # if near to integer binning factor use faster binning
#     #if abs(binfac - int(binfac)) < .05:
#     #    im2 = downscale_local_mean(im, (int(binfac), int(binfac)))
#     #else:
#     im2 = rescale(im, 1 / binfac, anti_aliasing=True, mode='constant', multichannel=False)
#     if original_size:
#         return resize(im2, im.shape, anti_aliasing=False, mode='constant')
#     else:
#         return im2

# ----------- TRIG

rads = math.pi / 180

def angle360(angle):
    if angle < 0:
        return angle + 360
    elif angle > 360:
        return angle - 360
    else:
        return angle

def angle_diff(t1, t2):
    # anticlockwise angle between 2 positions relative to 0 = positive x-axis
    return angle360(math.atan2(t2[1] - t1[1], t2[0] - t1[0]) / rads)

def p2cart(r, theta):
    return r * math.cos(rads*theta), r * math.sin(rads*theta)

class Polar:

    def __init__(self, origin=[0, 0], angle=0, radius=0):
        self.origin = origin
        self.angle = angle
        self.radius = radius

    def to_pos(self):
        return self.origin[0] + self.radius * math.cos(self.angle * rads), \
                self.origin[1] + self.radius * math.sin(self.angle * rads)

    def dist_to(self, pos):
        return Vector(self.to_pos()).distance(pos)

# ===========file/dir related

# note that if this doesn't quite work in all situations, there is a slightly modified
# approach in FileBrowser 
def home_folder():
    return os.path.expanduser('~')
    #if platform == 'win':
    #    return os.path.dirname(os.path.expanduser('~')) + os.path.sep
    #else:
    #    return os.path.expanduser('~') + os.path.sep

def add_if_not_exists(path):
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except FileExistsError as e:
            Logger.exception('Utils: add_if_not_exists')

def get_fits(path, time_order=False):
    # return a list of all fits files from path, possibly in modification time order
    names = glob.glob(os.path.join(path, '*.fit*')) + glob.glob(os.path.join(path, '*.FIT*'))
    if time_order:
        try:
            nms = {nm: os.path.getmtime(nm) for nm in names}
            names = [n for n, _ in sorted(nms.items(), key=operator.itemgetter(1))]
        except OSError as e:
            # if any issues getting time order just return in any order
            Logger.debug('Utils: get_fits problem getting mod time')
            names = glob.glob(os.path.join(path, '*.fit*')) + glob.glob(os.path.join(path, '*.FIT*'))
    return names

def dodgy_name(nm):
    chars = set(':/\\')
    return any((c in chars) for c in nm)
 
def purify_name(nm):
    # replace any chars that could cause problems with filesystem by spaces
    return nm.replace(':',' ').replace('/',' ').replace('\\',' ').strip()

def unique_dir_name(prefix, path, default_name='anon'):
    # Generate a unique folder name for folder 'prefix' adding datetime if necc

    if prefix is None:
        prefix = default_name
    if len(prefix) == 0:
        prefix = default_name

    if not(os.path.exists(os.path.join(path, prefix))):
        return prefix
    else:
        return '{:}_{:}'.format(prefix, datetime.now().strftime('%d_%b_%y_%H_%M_%S'))
