# Lever. Widget acting as circular scroll control around the eyepiece.

from kivy.properties import NumericProperty
from kivy.lang import Builder
from kivy.uix.label import Label
from kivy.metrics import dp

from jocular.utils.utils import angle_diff
from jocular.widgets.rotateables import RotateableCollideable

Builder.load_string('''

<Lever>:
    size_hint: None, None
    size: self.text_width, self.font_size + dp(2)
    text_size: self.size
    halign: 'center'
    color: self.color[0], self.color[1], self.color[2], app.dimmed_transp

    markup: True
    canvas.before:
        PushMatrix
        Translate:
            x: -self.width/2
            y: -self.height/2
        Rotate:
            angle: self.rotangle
            origin: self.center
        Color:
            rgba: self.background_color
        Rectangle:
            pos: self.pos
            size: self.size
        Color:
            rgba: self.marker_color
        Line:
            points: (self.x, self.center_y, self.x - self.marker_len, self.center_y) \
                if (self.angle < 90) or (self.angle > 270) else \
                (self.x + self.width, self.center_y, self.x + self.width + self.marker_len, self.center_y)
            width: dp(1)
    canvas.after:
        PopMatrix

''')

class Lever(Label, RotateableCollideable):

    value = NumericProperty(0.0)

    def __init__(self, value=0, values=[0, 1], angles=[0, 1], marker_len=None, 
        radial=True, springy=None, text_width=None, background_color=None, **kwargs):

        if marker_len is None:
            marker_len = dp(6)

        if text_width is None:
            text_width = dp(20)

        self.min_angle, self.max_angle = angles[0], angles[1]
        self.min_value, self.max_value = values[0], values[1]
        self.text_width = text_width
        self.background_color = background_color
        self.marker_len = marker_len
        self.springy = springy
        self.value =  value
        self.marker_color = self.background_color.copy()
        self.marker_color[-1] += .1
        super().__init__(radial=radial, **kwargs)
        self.register_event_type('on_value')
        self.selected = False
        self._k = (self.max_angle - self.min_angle) / (self.max_value - self.min_value)
        self.rotangle = self.value_to_angle(self.value)
        self.update(self.rotangle)

    def reset_value(self, v):
        # force update of lever position
        ang = self._k * (v - self.min_value) + self.min_angle
        self.update(ang)

    def value_to_angle(self, v):
        return self._k * (v - self.min_value) + self.min_angle

    def on_touch_down(self, touch):
        if self.collide_point(*touch.pos):
            self.selected = True
            return True
        return super().on_touch_down(touch)

    def on_touch_move(self, touch):
        if self.selected:
            self.update(angle_diff(self.origin, touch.pos))
 
    def on_touch_up(self, touch):
        self.selected = False
        if self.springy is not None:
            self.update(self.value_to_angle(self.springy))

    def update(self, angle, update_value=True):
        if update_value:
            mina, maxa = self.min_angle, self.max_angle
        
            # return if angle outside range
            if (angle < min(mina, maxa)) or (angle > max(mina, maxa)):
                return

            self.value =  float(((angle - mina) / self._k) + self.min_value)

            # limit value to min, max
            self.value = min(self.max_value, max(self.value, self.min_value))

            # convert value to angle in case it hasn't changed
            angle = self.value_to_angle(self.value)

        self.angle = angle
        self.on_angle()
        self.dispatch('on_value', self.value)

    def on_value(self, *args):
        pass

