# Rotated widgets (eg labels, buttons) which handle more complex collision-detection

from kivy.uix.label import Label
from kivy.uix.button import Button
from kivy.uix.widget import Widget
from kivy.properties import NumericProperty, ObjectProperty, ListProperty
from kivy.lang import Builder
from kivy.metrics import dp

from jocular.utils.utils import Polar

Builder.load_string('''

<RotLabel>:
    size_hint: None, None
    size: self.text_width, self.font_size + dp(2)
    text_size: self.size
    halign: 'center'
    markup: True
    color: app.text_color
    canvas.before:
        PushMatrix
        Translate:
            x: -self.width/2
            y: -self.height/2
        Rotate:
            angle: self.rotangle
            origin: self.center
        Color:
            rgba: self.background_color
        Rectangle:
            pos: self.pos
            size: self.size
    canvas.after:
        PopMatrix

<RotButton>:
    size_hint: None, None
    size: self.text_width, self.font_size + dp(2)
    text_size: self.size
    halign: 'center'
    color: self.color[0], self.color[1], self.color[2], app.dimmed_transp
    markup: True
    canvas.before:
        PushMatrix
        Translate:
            x: -self.width/2
            y: -self.height/2
        Rotate:
            angle: self.rotangle
            origin: self.center
        Color:
            rgba: self.background_color
        Rectangle:
            pos: self.pos
            size: self.size
    canvas.after:
        PopMatrix

''')

from kivy.graphics.transformation import Matrix
from math import radians

class Rotateable(Widget, Polar):

    # rotation angle in degrees relative to x-axis, measured anticlockwise
    rotangle = NumericProperty(0.0)

    deltas = {
        (1, True): 0, (1, False): -90,
        (2, True): -180, (2, False): -90,
        (3, True): -180, (3, False): 90,
        (4, True): 0, (4, False): 90
        }

    def __init__(self, radial=False, **kwargs):
        super().__init__(**kwargs)
        self.radial = radial
        self.on_angle()

    def on_angle(self, *args):
        q = min(4, (self.angle // 90) + 1)
        self.pos = self.to_pos()
        self.rotangle = self.angle + self.deltas[(q, self.radial)]

# rotated with collision detection
class RotateableCollideable(Rotateable):

    transform_inv = ObjectProperty(Matrix())

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def on_angle(self, *args):
        super().on_angle(*args)
        # transformation matrix
        tr = Matrix().translate(-self.width/2, -self.height/2, 0)
        rot = Matrix().rotate(-radians(self.rotangle), 0, 0, 1)
        # rotate about center
        t2 = Matrix().translate(self.center[0], self.center[1], 0)
        t2 = t2.multiply(rot)
        t2 = t2.multiply(Matrix().translate(-self.center[0], -self.center[1], 0))
        # multiply translate and rotate
        tr = tr.multiply(t2)
        # update inverse
        self.transform_inv = tr.inverse()
 
    def collide_point(self, x, y):
        # transform touchpos using the inverse transform of the rotation
        # takes about 150 microseconds
        x, y, _ = self.transform_inv.transform_point(x, y, 0)
        return self.x < x < (self.x + self.width) and self.y < y < (self.y + self.height)
 
# basic rotated label class
class RotLabel(Label, Rotateable):
    background_color = ListProperty([0, 0, 0, 0])
    color = ListProperty([.8, .8, .8, 0])
    text_width = NumericProperty(dp(20))

# basic rotated button class
class RotButton(Button, RotateableCollideable):
    background_color = ListProperty([0, 0, 0, 0])
    text_width = NumericProperty(dp(20))

# rotated toggle; actually, a multistate button that cycles thru alternatives
# acts as a normal button if only one value provided
class RotatedToggle(RotLabel, RotateableCollideable):

    def __init__(self, values=None, **kwargs):
        super().__init__(**kwargs)
        self.register_event_type('on_press')
        if values is None:
            self.values = []
        else:
            self.values = values
        if self.text not in self.values:
            self.values.append(self.text)  # default is a single state button

    def on_touch_down(self, touch):
        if super().collide_point(*touch.pos):
            # find location of self.text in self.values
            if len(self.values) > 1:
                loc = [i for i, x in enumerate(self.values) if x == self.text][0] + 1
                if loc == len(self.values):
                    loc = 0
                self.text = self.values[loc]
            self.dispatch('on_press')
            return True

    def on_press(self):
        pass
