# Table. Class representing sorteable, selectable table view associated with a json model.

import math
import operator
from datetime import datetime
import numpy as np

from kivy.logger import Logger
from kivy.core.window import Window
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.checkbox import CheckBox
from kivy.uix.button import Button
from kivy.uix.textinput import TextInput
from kivy.uix.label import Label
from kivy.lang import Builder
from kivy.properties import NumericProperty, StringProperty
from kivy.metrics import dp

from jocular.utils.utils import is_number

Builder.load_string('''

<TableFormat>:
    font_size: '16sp'
    margin: dp(5)
    padding: dp(5), dp(5)
    text_size: self.size
    color: .5, .5, .5, 1
    valign: 'middle'

<HeaderButton>:
    on_press: root.sort_column()
    browser: None
    background_color: .08, .08, .08, 1  # app.background_color

# row label
<TableLabel>:
    background_normal: ''
    height: dp(24)

# name column button
<TableButton>:
    column: ''
    background_normal: ''
    background_disabled_normal: ''
    background_color: 0, 0, 0, 0

# control button
<CButton>:
    size_hint: None, 1
    browser: None
    halign: 'center'
    background_normal: ''
    background_color: .06, .06, .06, 1

<SearchBarTextInput@TextInput>:
    background_color: .1, .1, .1, 1  # app.background_color
    foreground_color: .5, .5, .5, 1  # app.text_color
    multiline: False
    font_size: '16sp'

<SearchBar>:
    canvas.after:
        Color:
            rgba: app.line_color
        Line:
            points: self.x, self.y + dp(2), self.x + self.width, self.y + dp(2)
            width: 1
    size_hint: 1, None
    height: dp(32)
    font_size: '20sp'
    orientation: 'horizontal'

<TableHeader>:
    canvas.after:
        Color:
            rgba: app.line_color
        Line:
            points: self.x, self.y + dp(2), self.x + self.width, self.y + dp(2)
            width: dp(1)
    size_hint: 1, None
    height: dp(32)
    font_size: '20sp'
    orientation: 'horizontal'

<TableRow>:
    size_hint: 1, None
    height: dp(28)
    orientation: 'horizontal'

<Table>:
    size: dp(1100), dp(600)
    contents: _contents
    myheader: _header
    footer: _footer
    spacer: _spacer
    action: _action

    BoxLayout:
        orientation: 'vertical'
        pos: root.pos
        canvas.before:
            Color:
                rgba: .08, .08, .08, 1  # app.background_color 
            Rectangle:
                pos: self.pos
                size: self.size

        BoxLayout:
            id: _header
            size_hint: 1, None
            height: dp(64)
            orientation: 'vertical'

        BoxLayout:
            id: _allcontents
            orientation: 'horizontal'
            size_hint: 1, 1                

            BoxLayout:
                id: _contents
                size_hint: None, 1
                width: root.width - dp(30)
                orientation: 'vertical'

            Slider:
                size_hint: None, 1
                cursor_size: [dp(25), dp(25)]
                width: dp(30)
                disabled: root.n_matching < root.n_rows
                min: 0
                max: max(0, root.n_matching - root.n_rows)
                step: 1
                value: round(max(0, root.n_matching - root.n_rows) - root.current_row)
                on_value: root.current_row = round(max(0, root.n_matching - root.n_rows) - self.value)
                orientation: 'vertical'
                padding: dp(16)

        BoxLayout:
            id: _spacer
            size_hint: 1, None
            canvas.after:
                Color:
                    rgba: app.line_color
                Line:
                    points: self.x, self.y + self.height, self.x + self.width, self.y + self.height
                    width: 1
            size_hint: 1, None
            height: dp(4)
            orientation: 'horizontal'

        BoxLayout:
            id: _footer
            size_hint: 1, None
            height: dp(32)
            orientation: 'horizontal'
            
            TableLabel:
                size_hint: None, 1
                width: dp(340)
                halign: 'left'
                text: root.name
                font_size: '28sp'
                height: dp(32)

            TableLabel:
                size_hint: None, 1
                width: dp(200)
                text: 'showing {:d} of {:d}'.format(root.n_matching, root.n_rows_total)

            Label:
                size_hint: .2, 1

            CButton:
                size_hint: None, 1
                width: dp(100)
                text: 'select all'
                text_size: self.size
                on_press: root.select_all()
            CButton:
                size_hint: None, 1
                width: dp(110)
                text: 'deselect all'
                text_size: self.size
                on_press: root.deselect_all()

            CButton:
                id: _action
                size_hint: None, 1
                width: dp(200)
                text: ''

            Label:
                size_hint: .2, 1

''')

class TableFormat:
    pass

# header column button (controls sorting)
class HeaderButton(Button, TableFormat):

    def __init__(self, table=None, column_name=None, column_type=None, sort_type=None, **kwargs):
        super().__init__(**kwargs)
        self.column_name = column_name
        self.table = table
        self.column_type = column_type
        self.sort_type = sort_type

    def sort_column(self):
        self.table.current_sort = {
            'name': self.column_name,
            'type': self.column_type,
            'sort_type': self.sort_type
        }
        self.table.sort_column()

# ordinary label
class TableLabel(Label, TableFormat):
    pass

# control button at lower of screen
class CButton(Button, TableFormat):
    pass

# ordinary button
class TableButton(Button, TableFormat):
    pass

#-------------------------------------------------------------------------------------------
# table header line i.e. click-to-sort column headers

class TableHeader(BoxLayout):

    def __init__(self, table, **kwargs):
        super().__init__(**kwargs)
        self.table = table

        # selection column
        self.add_widget(HeaderButton(size_hint_x=None, width=dp(30), halign='center', 
            table=table, text=' ', column_name='Selected', column_type=str))

        for i, (colname, colprops) in enumerate(table.cols.items()):
            width = colprops.get('w', dp(100))

            self.add_widget(HeaderButton(
                text=colname,
                size_hint_x=width if width < 2 else None,
                width=width if width >= 2 else 1,
                halign='left' if (i == 0) else 'center',
                column_name=colprops.get('field', colname),
                column_type=colprops.get('type', str),
                sort_type=colprops.get('sort', 'normal'),
                table=table
                ))

        # close button
        w = CButton(text='close', size_hint_x=None, halign='right', width=dp(50))
        w.bind(on_press=self.table.hide)
        self.add_widget(w) 


    def resize(self, *args):
        pass

#-------------------------------------------------------------------------------------------
# a single row of the table

class TableRow(BoxLayout):

    def __init__(self, table, **kwargs):

        self.table = table
        super().__init__(**kwargs)

        # construct selector column
        w = CheckBox(size_hint_x=None, width=30)
        w.bind(active=self.select_row)
        self.add_widget(w)

        self.fields = {'Selected': w}

        for colname, colprops in self.table.cols.items():

            if 'action' in colprops:
                wtype = TableButton
            else:
                wtype = TableLabel

            width = colprops.get('w', dp(100))

            w = wtype(
                text='',
                size_hint_x=width if width < 2 else None,
                width=width if width >= 2 else 1,
                halign=colprops.get('align', 'center'),
                shorten=True,
                shorten_from='right')

            if 'action' in colprops:
                w.bind(on_press=self.row_clicked)  # add partial with column name

            self.add_widget(w)
            self.key = None

            self.fields[colprops.get('field', colname)] = w  # not sure about this


    def row_clicked(self, *args):
        # only supports one at present <-- need to fix this using partial
        for colname, colprops in self.table.cols.items():
            if 'action' in colprops:
                colprops['action'](self)
                return

    # prevent click thru to beneath
    def on_touch_down(self, touch):
        handled = super().on_touch_down(touch)
        if self.collide_point(*touch.pos):
            return True
        return handled

    def select_row(self, *args):
        if self.fields['Selected'].active:
            self.table.selected.add(self.key)
        elif self.key in self.table.selected:
            self.table.selected.remove(self.key)

    def update_row(self, row, key=None, empty=False):

        if not empty:
            self.key = key
            self.fields['Selected'].active = self.key in self.table.selected
            self.fields['Selected'].disabled = False
        else:
            self.fields['Selected'].active = False
            self.fields['Selected'].disabled = True

        for colname, colprops in self.table.cols.items():
            field = colprops.get('field', colname)  # speed this up
            if empty:
                self.fields[field].text = ''
            elif field in row:
                self.fields[field].text = str(row[field])
            else:
                self.fields[field].text = ''

#-------------------------------------------------------------------------------------------
# table search line i.e. column filters

class SearchBar(BoxLayout):

    def __init__(self, table, **kwargs):

        self.table = table
        super().__init__(**kwargs)

        self.sorted_column = 'Name'
        self.reverse_sort = {'Name': False}

        self.filters = []

        w = SearchFilter(text='', size_hint_x=None, width=30, table=table, column='Selected', column_type=str)
        self.add_widget(w)
        self.filters.append(w)

        for colname, colprops in self.table.cols.items():
            width = colprops.get('w', dp(100))

            w = SearchFilter(
                text='',
                size_hint_x=width if width < 2 else None,
                width=width if width >= 2 else 1,
                column=colprops.get('field', colname),
                table=table,
                val_fn=colprops.get('val_fn', None),
                column_type=colprops.get('type', str)
                )

            self.add_widget(w)
            self.filters.append(w)
 
        clear_button = CButton(text='clear', size_hint_x=None, width=dp(50), background_color=(.1, .1, .1, 1))
        self.add_widget(clear_button)
        clear_button.bind(on_press=self.clear_search)

    def clear_search(self, *args):
        for f in self.filters:
            f.text = ''
        self.table.search_results = list(self.table.data.keys())
        self.table.on_search_results()

    def resize(self, *args):
        pass

    def combine_results(self):
        l = [s.filtered for s in self.filters]
        self.table.search_results = list(set.intersection(*l)) # causes update
        self.table.on_search_results()

    def reapply_filters(self):
        # called whenever the underlying table data has changed
        for f in self.filters:
            f.has_had_text = False
            f.on_text(f, f.text)
        self.combine_results()

#-------------------------------------------------------------------------------------------
# an individual column filter

class SearchBarTextInput(TextInput):
    pass

class SearchFilter(SearchBarTextInput):

    def __init__(self, table=None, column=None, column_type=None, val_fn=None, **kwargs):

        self.table = table
        self.column = column
        self.column_type = column_type
        self.val_fn = val_fn
        self.has_had_text = False   # see comment in on_text for the why
        super().__init__(**kwargs)

        # current set that matches filter for this column
        self.filtered = set({})


    def on_text(self, instance, value):
        ''' Applies search filter when user types. We use a flag to indicate if 
        the search field has ever contained text. If so, when it becomes empty 
        is is necessary to update the results. Otherwise, we don't update on empty 
        to avoid multiple identical updates.
        '''

        value = value.strip()

        if not value:
            self.filtered = set(self.table.data.keys())
            if self.has_had_text:
                self.table.searchbar.combine_results()
            return

        try:

            objs = self.table.data
            col = self.column
            col_type = self.column_type

            if len(value) > 0:

                self.has_had_text = True

                # select column where it exists
                objs = {k: v[col] for k, v in objs.items() if col in v}

                if self.column_type == str:

                    objs = {k: v.lower() for k, v in objs.items()}
                    value = value.lower()

                    # use wildcard syntax, and = for exact match
                    if value[0] == '*':
                        if len(value) > 1:
                            if value[-1] == '*':
                                objs = {k: v for k, v in objs.items() if value[1:-1] in v}
                            else:
                                objs = {k: v for k, v in objs.items() if v.endswith(value[1:])}
                    elif value[0] == '=':
                        objs = {k: v for k, v in objs.items() if v == value[1:]}
                    else:
                        if value[-1] == '*':
                            value = value[:-1]
                        objs = {k: v for k, v in objs.items() if v.startswith(value)}

                elif self.column_type in [float, int]:

                    objs = {k: col_type(v) for k, v in objs.items()}
                   
                    # use interval syntax e.g. x, y or x, or y, 
                    col = self.column

                    if ',' not in value:
                        if is_number(value):
                            value = float(value)
                            if self.val_fn:
                                value = self.val_fn(value)
                            objs = {k: v for k, v in objs.items() if v <= value}

                    else:
                        parts = value.strip().split(',')
                        p1 = parts[0].strip()
                        p2 = parts[1].strip()
                        if is_number(p1):
                            p1 = float(p1)
                            if self.val_fn:
                                p1 = self.val_fn(p1)
                            objs = {k: v for k, v in objs.items() if v >= p1}
                        if is_number(p2):
                            p2 = float(p2)
                            if self.val_fn:
                                p2 = self.val_fn(p2)
                            objs = {k: v for k, v in objs.items() if v <= p2}

            self.filtered = set(objs.keys())

            # update search results by intersecting with current search results from other columns
            self.table.searchbar.combine_results()

        except Exception as e:
            pass

#-------------------------------------------------------------------------------------------
# Main Table class

class Table(FloatLayout):

    n_rows = NumericProperty(0)      # num visible rows
    n_matching = NumericProperty(0)  # number matching search/filter criteria
    current_row = NumericProperty(0)
    n_rows_total = NumericProperty(0)
    name =StringProperty('')

    def __init__(self, size=None, data=None, cols=None, name=None, actions=None, 
        update_on_show=True, on_show_method=None, on_hide_method=None, **kwargs):

        self.initialising = True  # to prevent redraw
        super().__init__(**kwargs)

        self.data = data
        self.n_rows_total = len(self.data)   # need to ensure this is kept updated
        self.cols = cols
        self.name = name
        self.actions = actions  # currently support just one action button
        self.update_on_show = update_on_show
        self.selected = set({})
        self.on_show_method = on_show_method
        self.on_hide_method = on_hide_method

        self.filters = {}
        self.reverse_sort = {}
        self.search_results = list(self.data.keys())

        self.initialising = False  # now allow redraw
        self.size = Window.size

    def redraw(self, *args):

        def make_rows():
            h = self.height - self.myheader.height - self.footer.height
            nrows = int(h / self.empty_row.height)
            spare = h - nrows * self.empty_row.height
            rows = []
            for i in range(0, nrows):
                r = TableRow(self)
                rows.append(r)
            return rows, len(rows), spare
           
        if self.initialising:
            return

        self.redrawing = True

        self.size = Window.size

        if not hasattr(self, 'header'):
            # generate header/searchbar components
            self.header = TableHeader(self)
            self.myheader.add_widget(self.header)
            self.searchbar = SearchBar(self)
            self.myheader.add_widget(self.searchbar)
            self.empty_row = TableRow(self)
            self.rows, self.n_rows, spare = make_rows()
        else: # remove rows
            for r in self.rows:
                self.contents.remove_widget(r)

        self.rows, self.n_rows, spare = make_rows()
        for r in self.rows:
            self.contents.add_widget(r)
        self.spacer.height = spare

        if self.actions:
            # just support one for now
            for k, v in self.actions.items():
                self.action.text = k
                self.action.bind(on_press=v)

        self.redrawing = False
        self.on_search_results()

    on_size = redraw

    def on_search_results(self, *args):

        if self.initialising: #  or self.redrawing:
            return

        self.current_row = 0
        self.n_matching = len(self.search_results)
        self.sort_column(reverse=False)
        self.update_display()

    def update(self, *args):
        # performs immediate update of table contents following a change
        # typically called on show and also from clients who make changes to data
        self.selected = {s for s in self.selected if s in self.data}
        self.n_rows_total = len(self.data)
        if hasattr(self, 'searchbar'):
            self.searchbar.reapply_filters()

    def update_display(self, *args): 
        # populate screen cells with current data

        last_row = min(self.n_matching, self.current_row + self.n_rows)
        results = self.search_results[self.current_row: last_row]

        for i, r in enumerate(results):
            # catch odd bug and log it
            if r in self.data:
                self.rows[i].update_row(self.data[r], key=r)
            else:
                Logger.debug('Table: cannot find {:} in table'.format(r))

        for j in range(last_row - self.current_row, self.n_rows):  # clear rest
            self.rows[j].update_row(self.empty_row, empty=True)

    def show(self, *args):

        if self.on_show_method:
            self.on_show_method()

        self.pos = 0, 0

        if self.parent:
            if self.size != self.parent.size:
                self.size = self.parent.size

        if self.update_on_show:
            self.update()

        else:
            self.update_display()

    def hide(self, *args):
        if self.on_hide_method:
            self.on_hide_method()
        self.y = -self.height

    def on_current_row(self, *args):
        self.update_display()

    def select_all(self, *args):
        self.selected = set(self.search_results)
        self.update_display()

    def deselect_all(self, *args):
        self.selected = set({})
        self.update_display()

    def select(self, to_select):
        self.selected = set(to_select)
        self.update_display()

    def get_selected(self):
        return self.selected

    def on_touch_down(self, touch):
        # generic method
        handled = super().on_touch_down(touch)
        if self.collide_point(*touch.pos):
            return True
        return handled

    def sort_column(self, reverse=True):

        def defaultsplit(x, default):
            a = x.split()
            if len(a) == 0:
                return '', default
            elif len(a) == 1:
                return a[0], default
            elif a[1].isdigit():
                return a[0], float(a[1])
            elif a[0].isdigit():
                return float(a[0]), a[1]
            else:
                return x, default

        def null_value(v):
            return (isinstance(v, str) and v.strip() == '') or (isinstance(v, float) and  math.isnan(v))

        if self.initialising or self.redrawing:
            return

        if not hasattr(self, 'current_sort'):
            return

        column_name = self.current_sort['name']
        column_type = self.current_sort['type']
        sort_type = self.current_sort['sort_type']

        # toggle sort direction
        if column_name in self.reverse_sort:
            if reverse:
                self.reverse_sort[column_name] = not self.reverse_sort[column_name]
        else:
            self.reverse_sort[column_name] = False

        # current filtered results
        current = np.array(self.search_results)

        # set up dict from position to value for current column, where column exists
        x = {i: self.data[k][column_name] for i, k in enumerate(current) if column_name in self.data[k]}

        # remove any null values that have crept in
        x = {i: v for i, v in x.items() if v}

        if not x:
            return

        # store indices that won't be sorted as they don't have a value
        blank_indices = list(set(np.arange(len(self.search_results))) - set(x.keys()))

        # catch specialised sorts first
        if type(sort_type) == dict:
            if 'catalog' in sort_type: 
                # catalog sort sorts first and second columns e.g. NGC 1
                vals = [defaultsplit(str(v), 0) for v in x.values()]
                cat_sort_index = np.lexsort(([v[1] for v in vals], [v[0] for v in vals]))
                sort_index = np.array(list(x.keys()))[cat_sort_index]

            elif 'DateFormat' in sort_type:
                fmt = sort_type['DateFormat']
                x = {k: datetime.strptime(v, fmt) for k, v in x.items()}
                sort_index = [i for i, j in sorted(x.items(), key=operator.itemgetter(1))]

        elif column_type in [str, float, int]:
            x = {k: column_type(v) for k, v in x.items()}
            sort_index = [i for i, j in sorted(x.items(), key=operator.itemgetter(1))]
 
        else:
            sort_index = np.argsort(x.values())

        # reverse if necessary
        if self.reverse_sort[column_name]:
            sort_index = sort_index[::-1]

        # append blank indices
        self.search_results = list(current[sort_index]) + list(current[blank_indices])

        self.update_display()

 
